import base64
from binascii import hexlify
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
import json
from kvirt import common
from kvirt.common import error, pprint, warning, sdn_ip
from kvirt.defaults import UBUNTUS, METADATA_FIELDS
from kvirt.providers.vsphere.helpers import find, collectproperties, findvm, createfolder, changecd, convert, waitForMe
from kvirt.providers.vsphere.helpers import createscsispec, creatediskspec, createdvsnicspec, createclonespec
from kvirt.providers.vsphere.helpers import createnicspec, createisospec, createcdspec, deletedirectory
from kvirt.providers.vsphere.helpers import dssize, keep_lease_alive
from kvirt.providers.vsphere.helpers import create_filter_spec, get_all_obj, convert_properties, findvm2, findvmdc
import os
from pyVmomi import vim, vmodl
from pyVim import connect
import random
import re
import ssl
from ssl import _create_unverified_context, get_server_certificate
import sys
from shutil import which
from subprocess import call
import tarfile
from tempfile import TemporaryDirectory
from threading import Thread
import time
import urllib.request
from uuid import UUID
import webbrowser
from zipfile import ZipFile


class Ksphere:
    def __init__(self, host, user, password, datacenter, cluster, debug=False, isofolder=None,
                 filtervms=False, filteruser=False, filtertag=None, category='kcli', basefolder=None, dvs=True,
                 import_network='VM Network', timeout=3600, force_pool=False, restricted=False, serial=False):
        if timeout < 1:
            smart_stub = connect.SmartStubAdapter(host=host, port=443, sslContext=_create_unverified_context(),
                                                  connectionPoolTimeout=0)
            session_stub = connect.VimSessionOrientedStub(smart_stub,
                                                          connect.VimSessionOrientedStub.makeUserLoginMethod(user,
                                                                                                             password))
            si = vim.ServiceInstance('ServiceInstance', session_stub)
        elif hasattr(connect, 'SmartConnectNoSSL'):
            si = connect.SmartConnectNoSSL(host=host, port=443, user=user, pwd=password,
                                           connectionPoolTimeout=timeout)
        else:
            si = connect.SmartConnect(host=host, port=443, user=user, pwd=password, disableSslCertValidation=True,
                                      connectionPoolTimeout=timeout)
        self.conn = si
        self.si = si
        self.vcip = host
        self.category = category
        self.url = f"https://{user}:{password}@{host}/sdk"
        self.user = user
        self.password = password
        self.rootFolder = si.content.rootFolder
        self.dc = find(si, self.rootFolder, vim.Datacenter, datacenter)
        self.macaddr = []
        self.clu = cluster
        self.isofolder = isofolder
        self.filtervms = filtervms
        self.filtervms = filtervms
        self.filteruser = filteruser
        self.filtertag = filtertag
        self.debug = debug
        self.networks = []
        self.dvs = dvs
        self.portgs = {}
        self.esx = '.' in cluster and datacenter == 'ha-datacenter'
        self.restricted = restricted or self.esx
        self.serial = serial
        self.import_network = import_network
        self.force_pool = force_pool
        if basefolder is not None:
            if '/' in basefolder:
                topfolder = os.path.dirname(basefolder)
                vmFolder = find(si, self.dc.vmFolder, vim.Folder, topfolder)
                if vmFolder is None:
                    error(f"Couldnt find topfolder {topfolder}")
                    self.conn = None
                    return
                lowFolder = os.path.basename(basefolder)
            else:
                vmFolder = self.dc.vmFolder
                lowFolder = basefolder
            self.basefolder = find(si, vmFolder, vim.Folder, lowFolder)
            if self.basefolder is None:
                if not self.restricted:
                    try:
                        createfolder(si, vmFolder, lowFolder)
                        self.basefolder = find(si, vmFolder, vim.Folder, lowFolder)
                    except Exception as e:
                        error(f"Couldnt create basefolder {basefolder}. Hit {e}")
                        self.conn = None
                        return
                else:
                    error(f"Couldnt find basefolder {basefolder}")
                    self.conn = None
                    return
        else:
            self.basefolder = self.dc.vmFolder

    def set_networks(self):
        si = self.si
        view = si.content.viewManager.CreateContainerView(self.rootFolder, [vim.Network], True)
        netlist = collectproperties(si, view=view, objtype=vim.Network, pathset=['name'], includemors=True)
        self.networks = [o['obj'].name for o in netlist]
        if self.dvs:
            portgs = {}
            o = si.content.viewManager.CreateContainerView(self.rootFolder, [vim.DistributedVirtualSwitch], True)
            dvnetworks = o.view
            o.Destroy()
            for dvnetw in dvnetworks:
                uuid = dvnetw.uuid
                for portg in dvnetw.portgroup:
                    portgs[portg.name] = [uuid, portg.key]
            self.portgs = portgs

    def close(self):
        self.si.content.sessionManager.Logout()

    def exists(self, name):
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        return findvm(si, vmFolder, name) is not None

    def net_exists(self, name):
        print("not implemented")

    def create(self, name, virttype=None, profile='kvirt', flavor=None, plan='kvirt', cpumodel='host-model',
               cpuflags=[], cpupinning=[], numcpus=2, memory=512, guestid='centos7_64Guest', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True, diskinterface='virtio', nets=['default'], iso=None,
               vnc=True, cloudinit=True, reserveip=False, reservedns=False, reservehost=False, start=True, keys=[],
               cmds=[], ips=None, netmasks=None, gateway=None, nested=True, dns=None, domain=None, tunnel=False,
               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[],
               cmdline=None, placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False,
               numamode=None, numa=[], pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[],
               vmuser=None, guestagent=True):
        if not self.esx and self.exists(name):
            return {'result': 'failure', 'reason': f"VM {name} already exists"}
        dc = self.dc
        vmFolder = dc.vmFolder
        diskmode = 'persistent'
        default_diskinterface = diskinterface
        default_diskthin = diskthin
        default_disksize = disksize
        default_pool = pool
        memory = int(memory)
        numcpus = int(numcpus)
        si = self.si
        basefolder = self.basefolder
        restricted = self.restricted
        rootFolder = self.rootFolder
        basefolder = self.basefolder
        cluster = overrides.get('cluster')
        vmfolder = basefolder
        if not restricted:
            if cluster is not None:
                createfolder(si, basefolder, cluster)
                vmfolder = find(si, basefolder, vim.Folder, cluster)
            elif plan != 'kvirt':
                createfolder(si, basefolder, plan)
                vmfolder = find(si, basefolder, vim.Folder, plan)
        si = self.si
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        if 'resourcepool' in overrides:
            resourcepoolname = overrides['resourcepool']
            resourcepool = find(si, rootFolder, vim.ResourcePool, resourcepoolname)
            if resourcepool is None:
                return {'result': 'failure', 'reason': f"Resourcepool {resourcepoolname} not found"}
        else:
            resourcepool = clu.resourcePool
        if image is not None:
            imagepool = image.split('/')[0] if '/' in image else None
            image = os.path.basename(image)
            clonespec = createclonespec(resourcepool)
            rootFolder = self.rootFolder
            imageobj, imagedc = findvmdc(si, rootFolder, image, self.dc)
            if self.esx:
                imagepool = pool
                imagedc = self.dc.name
            elif imageobj is None:
                return {'result': 'failure', 'reason': f"Image {image} not found"}
            datastores = self._datastores_datacenters()
            if os.path.basename(datastores[pool]) != self.dc.name:
                return {'result': 'failure', 'reason': f"Pool {pool} doesn't belong to Datacenter {self.dc.name}"}
            if imagepool is None:
                devices = imageobj.config.hardware.device
                for number, dev in enumerate(devices):
                    if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
                        imagepool = dev.backing.datastore.name
            if imagedc != self.dc.name or (overrides.get('force_pool', self.force_pool) and imagepool != pool):
                warning(f"Vm {name} will be relocated from pool {imagepool} to {pool}")
                relospec = vim.vm.RelocateSpec()
                relospec.datastore = find(si, rootFolder, vim.Datastore, pool)
                relospec.pool = resourcepool
                clonespec.location = relospec
            confspec = vim.vm.ConfigSpec()
            confspec.flags = vim.vm.FlagInfo()
            confspec.flags.diskUuidEnabled = True
            confspec.annotation = name
            if memory % 1024 != 0:
                warning("Rounding up memory to be multiple of 1024")
                memory += (1024 - memory) % 1024
            confspec.memoryMB = memory
            confspec.memoryHotAddEnabled = memoryhotplug
            cores = overrides.get('cores')
            sockets, cores = overrides.get('sockets', 1), overrides.get('cores')
            if cores is not None and isinstance(cores, int) and isinstance(sockets, int):
                confspec.numCoresPerSocket = cores
                numcpus = cores * sockets
            threads = overrides.get('threads')
            if threads is not None and isinstance(threads, int):
                confspec.simultaneousThreads = threads
            confspec.numCPUs = numcpus
            confspec.cpuHotAddEnabled = cpuhotplug
            confspec.cpuHotRemoveEnabled = cpuhotplug
            extraconfig = []
            for entry in [field for field in metadata if field in METADATA_FIELDS]:
                opt = vim.option.OptionValue()
                opt.key = entry
                opt.value = metadata[entry]
                extraconfig.append(opt)
            if tags:
                opt = vim.option.OptionValue()
                opt.key = 'tags'
                opt.value = ','.join(tags)
                extraconfig.append(opt)
            clonespec.config = confspec
            clonespec.powerOn = False
            isofolder = self.isofolder if self.isofolder is not None else f"[{default_pool}]/{name}"
            cloudinitiso = f"{isofolder}/{name}.ISO"
            combustion = common.needs_combustion(image)
            if cloudinit:
                if image is not None and common.needs_ignition(image):
                    version = common.ignition_version(image)
                    meta, netdata = '', None
                    ignitiondata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                                   domain=domain, files=files, enableroot=enableroot,
                                                   overrides=overrides, version=version, plan=plan, image=image,
                                                   vmuser=vmuser)
                    if combustion:
                        userdata, meta, netdata = ignitiondata, '', None
                    else:
                        cloudinitiso = None
                        ignitionopt = vim.option.OptionValue()
                        ignitionopt.key = 'guestinfo.ignition.config.data'
                        ignitionopt.value = base64.b64encode(ignitiondata.encode()).decode()
                        encodingopt = vim.option.OptionValue()
                        encodingopt.key = 'guestinfo.ignition.config.data.encoding'
                        encodingopt.value = 'base64'
                        extraconfig.extend([ignitionopt, encodingopt])
                else:
                    gcmds = []
                    if image is not None and 'cos' not in image and 'fedora-coreos' not in image:
                        lower = image.lower()
                        if lower.startswith('fedora') or lower.startswith('rhel') or lower.startswith('centos'):
                            gcmds.append('yum -y install open-vm-tools')
                        elif lower.startswith('debian') or [x for x in UBUNTUS if x in lower] or 'ubuntu' in lower:
                            gcmds.append('apt-get update')
                            gcmds.append('apt-get -f install open-vm-tools')
                        gcmds.append('systemctl enable --now vmtoolsd')
                    index = 0
                    if image is not None and image.startswith('rhel'):
                        subindex = [i for i, value in enumerate(cmds) if value.startswith('subscription-manager')]
                        if subindex:
                            index = subindex.pop() + 1
                    cmds = cmds[:index] + gcmds + cmds[index:]
                    userdata, meta, netdata = common.cloudinit(name=name, keys=keys, cmds=cmds, nets=nets,
                                                               gateway=gateway, dns=dns, domain=domain,
                                                               files=files, enableroot=enableroot, overrides=overrides,
                                                               storemetadata=storemetadata, machine='vsphere',
                                                               image=image, vmuser=vmuser)
            for key in overrides:
                if key.startswith('guestinfo.'):
                    guestopt = vim.option.OptionValue()
                    guestopt.key = key
                    guestopt.value = overrides[key]
                    extraconfig.append(guestopt)
            confspec.extraConfig = extraconfig
            if not self.esx:
                t = imageobj.CloneVM_Task(folder=vmfolder, name=name, spec=clonespec)
            else:
                vm = findvm(si, vmFolder, name)
                t = vm.ReconfigVM_Task(confspec)
            waitForMe(t)
            if cloudinit and cloudinitiso is not None:
                with TemporaryDirectory() as tmpdir:
                    if combustion:
                        cmdsdata = common.process_combustion_cmds(cmds, overrides)
                        if cmdsdata != '':
                            with open(f'{tmpdir}/combustion_script', 'w') as combustionfile:
                                combustionfile.write(cmdsdata)
                    common.make_iso(name, tmpdir, userdata, meta, netdata, openstack=False, combustion=combustion)
                    cloudinitisofile = f"{tmpdir}/{name}.ISO"
                    isopool = default_pool
                    isofolder = None
                    if self.isofolder is not None:
                        isofolder = self.isofolder.split('/')
                        isopool = re.sub(r"[\[\]]", '', isofolder[0])
                        isofolder = isofolder[1]
                    self._uploadimage(isopool, cloudinitisofile, name, isofolder=isofolder)
                vm = findvm(si, vmFolder, name)
                c = changecd(self.si, vm, cloudinitiso)
                waitForMe(c)
        datastores = {}
        confspec = vim.vm.ConfigSpec()
        confspec.name = name
        confspec.annotation = name
        confspec.memoryMB = memory
        confspec.numCPUs = numcpus
        if 'uuid' in overrides:
            uuid = str(overrides['uuid'])
            try:
                UUID(uuid)
                confspec.uuid = uuid
            except:
                warning(f"couldn't use {uuid} as uuid")
        confspec.flags = vim.vm.FlagInfo()
        confspec.flags.diskUuidEnabled = True
        confspec.extraConfig = []
        for entry in [field for field in metadata if field in METADATA_FIELDS]:
            opt = vim.option.OptionValue()
            opt.key = entry
            opt.value = metadata[entry]
            confspec.extraConfig.append(opt)
        confspec.guestId = 'centos7_64Guest'
        vmfi = vim.vm.FileInfo()
        filename = "[" + default_pool + "]"
        vmfi.vmPathName = filename
        confspec.files = vmfi
        if vnc:
            vncport = random.randint(5900, 7000)
            opt1 = vim.option.OptionValue()
            opt1.key = 'RemoteDisplay.vnc.port'
            opt1.value = vncport
            opt2 = vim.option.OptionValue()
            opt2.key = 'RemoteDisplay.vnc.enabled'
            opt2.value = "TRUE"
            confspec.extraConfig.extend([opt1, opt2])
        if image is None:
            t = vmfolder.CreateVM_Task(confspec, resourcepool)
            waitForMe(t)
        vm = find(si, dc.vmFolder, vim.VirtualMachine, name)
        currentdevices = vm.config.hardware.device
        currentdisks = [d for d in currentdevices if isinstance(d, vim.vm.device.VirtualDisk)]
        currentnics = [d for d in currentdevices if isinstance(d, vim.vm.device.VirtualEthernetCard)]
        confspec = vim.vm.ConfigSpec()
        devconfspec = []
        unit_number = 1
        for index, disk in enumerate(disks):
            diskuuid = None
            if disk is None:
                disksize = default_disksize
                diskthin = default_diskthin
                diskinterface = default_diskinterface
                diskpool = default_pool
            elif isinstance(disk, int):
                disksize = disk
                diskthin = default_diskthin
                diskinterface = default_diskinterface
                diskpool = default_pool
            elif isinstance(disk, str) and disk.isdigit():
                disksize = int(disk)
                diskthin = default_diskthin
                diskinterface = default_diskinterface
                diskpool = default_pool
            elif isinstance(disk, dict):
                disksize = disk.get('size', default_disksize)
                diskthin = disk.get('thin', default_diskthin)
                diskinterface = disk.get('interface', default_diskinterface)
                diskpool = disk.get('pool', default_pool)
                diskuuid = disk.get('uuid') or disk.get('wwn')
                if diskuuid is not None:
                    try:
                        UUID(diskuuid)
                    except:
                        warning(f"{diskuuid} it not a valid disk uuid")
                        diskuuid = None
            if index < len(currentdisks) and image is not None:
                currentdisk = currentdisks[index]
                currentsize = convert(1000 * currentdisk.capacityInKB, GB=False)
                if int(currentsize) < disksize:
                    pprint(f"Waiting for image disk {index} to be resized")
                    currentdisk.capacityInKB = disksize * 1048576
                    diskspec = vim.vm.ConfigSpec()
                    diskspec = vim.vm.device.VirtualDeviceSpec(device=currentdisk, operation="edit")
                    devconfspec.append(diskspec)
                continue
            disksize = disksize * 1048576
            if diskpool not in datastores:
                datastore = find(si, rootFolder, vim.Datastore, diskpool)
                if not datastore:
                    return {'result': 'failure', 'reason': f"Pool {diskpool} not found"}
                else:
                    datastores[diskpool] = datastore
            if index == 0:
                scsispec = createscsispec()
                devconfspec.append(scsispec)
            if unit_number == 7:
                unit_number = 8
            diskspec = creatediskspec(unit_number, disksize, datastore, diskmode, diskthin, diskuuid)
            devconfspec.append(diskspec)
            unit_number += 1
        for index, _ in enumerate(disks):
            confspec.extraConfig.append(vim.option.OptionValue(key=f'scsi0:{index}.enableUUID', value='TRUE'))
        # NICSPEC
        if not self.networks:
            self.set_networks()
        macs = {}
        for index, net in enumerate(nets):
            netname = net['name'] if isinstance(net, dict) else net
            if isinstance(net, dict) and 'mac' in net:
                macs[index] = net['mac']
            if netname == 'default':
                if image is not None:
                    continue
                else:
                    netname = 'VM Network'
            if index < len(currentnics):
                currentnetwork = None
                currentnic = currentnics[index]
                try:
                    currentnetwork = currentnic.backing.deviceName
                except:
                    currentswitchuuid = currentnic.backing.port.switchUuid
                    currentportgroupkey = currentnic.backing.port.portgroupKey
                    for dvsnet in self.portgs:
                        if self.portgs[dvsnet][0] == currentswitchuuid and\
                                self.portgs[dvsnet][1] == currentportgroupkey:
                            currentnetwork = dvsnet
                if currentnetwork is None:
                    warning(f"Couldn't figure out network associated to nic {index}")
                elif currentnetwork != netname:
                    if netname in self.portgs:
                        switchuuid = self.portgs[netname][0]
                        portgroupkey = self.portgs[netname][1]
                        currentnic.backing.port.switchUuid = switchuuid
                        currentnic.backing.port.portgroupKey = portgroupkey
                        currentnic.backing.port.portKey = None
                    elif netname in self.networks:
                        currentnic.backing.deviceName = netname
                    else:
                        return {'result': 'failure', 'reason': f"Invalid network {netname}"}
                    nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
                    devconfspec.append(nicspec)
                continue
            nicname = f'Network Adapter {index + 1}'
            nictype = net['type'] if isinstance(net, dict) and 'type' in net else None
            if netname in self.portgs:
                switchuuid = self.portgs[netname][0]
                portgroupkey = self.portgs[netname][1]
                nicspec = createdvsnicspec(nicname, netname, switchuuid, portgroupkey, nictype=nictype)
            elif netname in self.networks:
                nicspec = createnicspec(nicname, netname, nictype=nictype)
            else:
                return {'result': 'failure', 'reason': f"Invalid network {netname}"}
            devconfspec.append(nicspec)
        need_cdrom = iso is not None or not cloudinit or image is None or common.needs_ignition(image)
        if iso is not None:
            matchingisos = [i for i in self._getisos() if i.endswith(iso)]
            if matchingisos:
                iso = matchingisos[0]
            else:
                return {'result': 'failure', 'reason': f"Iso {iso} not found"}
        if need_cdrom:
            cdspec = createcdspec() if iso is None and self.esx else createisospec(iso)
            devconfspec.append(cdspec)
        serial = overrides.get('serial', self.serial)
        if serial:
            serialdevice = vim.vm.device.VirtualSerialPort()
            backing = vim.vm.device.VirtualSerialPort.URIBackingInfo()
            backing.serviceURI = f'tcp://:{common.get_free_port()}'
            backing.direction = 'server'
            serialdevice.backing = backing
            serialdevice.key = len(devconfspec) + 1
            serialspec = vim.vm.device.VirtualDeviceSpec(device=serialdevice, operation="add")
            devconfspec.append(serialspec)
        confspec.deviceChange = devconfspec
        if nested:
            confspec.nestedHVEnabled = True
        uefi = overrides.get('uefi', False)
        uefi_legacy = overrides.get('uefi_legacy', False)
        secureboot = overrides.get('secureboot', False)
        if secureboot or uefi or uefi_legacy:
            confspec.firmware = 'efi'
            if secureboot:
                confspec.bootOptions = vim.vm.BootOptions(efiSecureBootEnabled=True)
        confspec.vAppConfigRemoved = True
        t = vm.Reconfigure(confspec)
        waitForMe(t)
        if overrides.get('boot_order', False):
            key_disks = [d.key for d in vm.config.hardware.device if isinstance(d, vim.vm.device.VirtualDisk)]
            boot_disks = [vim.vm.BootOptions.BootableDiskDevice(deviceKey=key) for key in key_disks]
            confspec = vim.vm.ConfigSpec()
            confspec.bootOptions = vim.vm.BootOptions(bootOrder=boot_disks)
            t = vm.Reconfigure(confspec)
            waitForMe(t)
        if macs:
            self.set_macs(name, macs)
        if 'vmgroup' in overrides:
            vmgroup = overrides['vmgroup']
            vmgroups = {}
            hostgroups = {}
            for group in clu.configurationEx.group:
                if hasattr(group, 'vm'):
                    vmgroups[group.name] = group
                else:
                    hostgroups[group.name] = group
            if vmgroup in vmgroups:
                vmgroup = vmgroups[vmgroup]
                vmgroup.vm.append(vm)
                vmgroupspec = vim.cluster.GroupSpec(info=vmgroup, operation='edit')
                groups_spec = vim.cluster.ConfigSpecEx(groupSpec=[vmgroupspec])
                t = clu.ReconfigureEx(groups_spec, modify=True)
                waitForMe(t)
            else:
                vmgroup = vim.cluster.VmGroup(name=vmgroup, vm=[vm])
                vmgroupspec = vim.cluster.GroupSpec(info=vmgroup, operation='add')
                groups_spec = vim.cluster.ConfigSpecEx(groupSpec=[vmgroupspec])
                t = clu.ReconfigureEx(groups_spec, modify=True)
                waitForMe(t)
            if 'hostgroup' in overrides and 'hostrule' in overrides:
                hostgroup = overrides['hostgroup']
                hostrule = overrides['hostrule']
                if hostgroup not in hostgroups:
                    msg = f"Hostgroup {hostgroup} not found. It needs to exist prior to vm's creation"
                    return {'result': 'failure', 'reason': msg}
                else:
                    hostgroup = hostgroups[hostgroup]
                    vmhostrulefound = False
                    for vmhostrule in clu.configurationEx.rule:
                        if vmhostrule.name == hostrule:
                            vmhostrulefound = True
                            break
                    if not vmhostrulefound:
                        pprint(f"Creating vmhost rule {hostrule}")
                        rule_obj = vim.cluster.VmHostRuleInfo(vmGroupName=hostrule, affineHostGroupName=hostgroup.name,
                                                              name=vmgroup.name, enabled=True, mandatory=True)
                        rulespec = vim.cluster.RuleSpec(info=rule_obj, operation='add')
                        groups_spec = vim.cluster.ConfigSpecEx(rulesSpec=[rulespec])
                        t = clu.ReconfigureEx(groups_spec, modify=True)
                        waitForMe(t)
        antipeers = overrides.get('antipeers', [])
        if antipeers and antipeers[-1] == name:
            antipeers_rule = '-'.join(antipeers)
            pprint(f"Creating anti affinity rule {antipeers_rule}")
            vms = []
            for member in antipeers:
                vm = findvm(si, vmFolder, member)
                if vm is None:
                    error(f"VM {member} not found")
                else:
                    vms.append(vm)
            if len(vms) > 1:
                rule_obj = vim.cluster.AntiAffinityRuleSpec(vm=vms, enabled=True, mandatory=True, name=antipeers_rule)
                rulespec = vim.cluster.RuleSpec(info=rule_obj, operation='add')
                groups_spec = vim.cluster.ConfigSpecEx(rulesSpec=[rulespec])
                t = clu.ReconfigureEx(groups_spec, modify=True)
                waitForMe(t)
        if start:
            t = vm.PowerOnVM_Task(None)
            waitForMe(t)
        if tags:
            pprint("Assigning tags")
            from kvirt.providers.vsphere.tagging import KsphereTag
            ktag = KsphereTag(self.vcip, self.user, self.password)
            categories = {}
            tags_ids = []
            for entry in tags:
                if isinstance(entry, str) and '=' in entry:
                    category, tag = entry.split('=')
                elif isinstance(entry, dict) and len(list(entry.keys())) == 1:
                    category, tag = list(entry.keys())[0], list(entry.values())[0]
                else:
                    category, tag = self.category, entry
                if category not in categories:
                    category_id = ktag.get_category_id(category)
                    if category_id is None:
                        category_id = ktag.create_category(category)
                    categories[category] = category_id
                category_id = categories[category]
                tag_id = ktag.get_tag_id(tag)
                if tag_id is None:
                    tag_id = ktag.create_tag(category_id, tag)
                tags_ids.append(tag_id)
            if tags_ids:
                vm_id = vm._moId
                ktag.add_tags(vm_id, tags_ids)
        return {'result': 'success'}

    def start(self, name):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        runtime = info['runtime']
        if runtime.powerState == "poweredOff":
            t = vm.PowerOnVM_Task(None)
            waitForMe(t)
        return {'result': 'success'}

    def stop(self, name, soft=False):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        runtime = info['runtime']
        if runtime.powerState == "poweredOn":
            t = vm.PowerOffVM_Task()
            waitForMe(t)
        return {'result': 'success'}

    def restart(self, name):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        try:
            vm.RebootGuest()
        except:
            vm.ResetVM_Task()
        return {'result': 'success'}

    def status(self, name):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        runtime = info['runtime']
        return runtime.powerState if vm is not None else ''

    def delete(self, name, snapshots=False):
        si = self.si
        dc = self.dc
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        summary = info['summary']
        config = info['config']
        runtime = info['runtime']
        plan, image, kube = 'kvirt', None, None
        vmpath = summary.config.vmPathName.replace(f'/{name}.vmx', '')
        if config is not None:
            for entry in config.extraConfig:
                if entry.key == 'image':
                    image = entry.value
                if entry.key == 'plan':
                    plan = entry.value
                if entry.key == 'kube':
                    kube = entry.value
        if runtime.powerState == "poweredOn":
            t = vm.PowerOffVM_Task()
            waitForMe(t)
        t = vm.Destroy_Task()
        waitForMe(t)
        if image is not None and 'coreos' not in image and 'rhcos' not in image and\
                'fcos' not in image and vmpath.endswith(name):
            isopath = f"{self.isofolder}/{name}.ISO" if self.isofolder is not None else vmpath
            try:
                deletedirectory(si, dc, isopath)
            except:
                pass
        if kube is not None:
            clusterfolder = find(si, vmFolder, vim.Folder, kube)
            if clusterfolder is not None and len(clusterfolder.childEntity) == 0:
                clusterfolder.Destroy()
        elif plan != 'kvirt':
            planfolder = find(si, vmFolder, vim.Folder, plan)
            if planfolder is not None and len(planfolder.childEntity) == 0:
                try:
                    planfolder.Destroy()
                except Exception as e:
                    error(f"Couldn't delete plan folder {plan}. Hit {e}")
        return {'result': 'success'}

    def serialconsole(self, name, web=False):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            print(f"VM {name} not found")
            return
        runtime = info['runtime']
        config = info['config']
        if runtime.powerState == "poweredOff":
            print("VM down")
        serialfound = False
        devices = config.hardware.device
        for dev in devices:
            if type(dev).__name__ == 'vim.vm.device.VirtualSerialPort':
                serialfound = True
                serialport = dev.backing.serviceURI.split(':')[2]
                break
        if serialfound:
            host = runtime.host.name
            url = f"{host} {serialport}"
            consolecommand = f"nc {url}"
            if web:
                return url
            if self.debug or os.path.exists("/i_am_a_container"):
                print(consolecommand)
            if not os.path.exists("/i_am_a_container"):
                call(consolecommand, shell=True)

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        si = self.si
        vcip = self.vcip
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            print(f"VM {name} not found")
            return
        runtime = info['runtime']
        config = info['config']
        if runtime.powerState == "poweredOff":
            print("VM down")
            return
        extraconfig = config.extraConfig
        vncfound = False
        for extra in extraconfig:
            key, value = extra.key, extra.value
            if 'vnc' in key and 'port' in key:
                vncfound = True
                vncport = value
                break
            else:
                continue
        if vncfound:
            consolecommand = ''
            host = runtime.host.name
            if tunnel and tunnelhost is not None:
                localport = common.get_free_port()
                consolecommand += "ssh -o LogLevel=QUIET -f -p "
                consolecommand += f"{tunnelport} -L {localport}:{host}:{vncport} {tunneluser}@{tunnelhost} sleep 10;"
                host = '127.0.0.1'
                vncport = localport
            url = f"vnc://{host}:{vncport}"
            if os.path.exists('/Applications'):
                if os.path.exists('/Applications/VNC Viewer.app'):
                    consolecommand += f"open -a 'VNC Viewer' --args {url.replace('vnc://', '')} &"
                else:
                    consolecommand += f"open -a 'Screen Sharing' {url} &"
            else:
                consolecommand += f"remote-viewer {url} &"
            if web:
                return url
            if self.debug or os.path.exists("/i_am_a_container"):
                print(consolecommand)
            if not os.path.exists("/i_am_a_container"):
                os.popen(consolecommand)
        elif self.esx:
            vmnumber = None
            devices = config.hardware.device
            for dev in devices:
                if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
                    vmnumber = dev.diskObjectId.split('-')[0]
                    break
            if vmnumber is not None:
                vmurl = f"https://{self.vcip}/ui/#/console/{vmnumber}"
                if self.debug or os.path.exists("/i_am_a_container"):
                    msg = f"Open the following url:\n{vmurl}" if os.path.exists("/i_am_a_container") else vmurl
                    pprint(msg)
                else:
                    pprint(f"Opening url {vmurl}")
                    webbrowser.open(vmurl, new=2, autoraise=True)
        else:
            content = si.RetrieveContent()
            sgid = content.about.instanceUuid
            cert = get_server_certificate((self.vcip, 443))
            cert_deserialize = x509.load_pem_x509_certificate(cert.encode(), default_backend())
            finger_print = hexlify(cert_deserialize.fingerprint(hashes.SHA1())).decode('utf-8')
            sha = ":".join([finger_print[i: i + 2] for i in range(0, len(finger_print), 2)])
            vcenter_data = content.setting
            vcenter_settings = vcenter_data.setting
            for item in vcenter_settings:
                key = getattr(item, 'key')
                if key == 'VirtualCenter.FQDN':
                    fqdn = getattr(item, 'value')
            sessionmanager = si.content.sessionManager
            session = sessionmanager.AcquireCloneTicket()
            vmid = vm._moId
            vmurl = f"https://{vcip}/ui/webconsole.html?"
            vmurl += f"vmId={vmid}&vmName={name}&serverGuid={sgid}&host={fqdn}&sessionTicket={session}&thumbprint={sha}"
            if web:
                return vmurl
            if self.debug or os.path.exists("/i_am_a_container"):
                msg = f"Open the following url:\n{vmurl}" if os.path.exists("/i_am_a_container") else vmurl
                pprint(msg)
            else:
                pprint(f"Opening url {vmurl}")
                webbrowser.open(vmurl, new=2, autoraise=True)

    def info(self, name, output='plain', fields=[], values=False, vm=None, debug=False):
        translation = {'poweredOff': 'down', 'poweredOn': 'up', 'suspended': 'suspended'}
        image = None
        yamlinfo = {}
        si = self.si
        if vm is None:
            listinfo = False
            vmFolder = self.basefolder
            obj, vm = findvm2(si, vmFolder, name)
            if vm is None:
                error(f"VM {name} not found")
                return {}
        else:
            listinfo = True
        summary = vm['summary']
        config = vm['config']
        runtime = vm['runtime']
        guest = vm['guest']
        yamlinfo['name'] = name
        yamlinfo['id'] = summary.config.instanceUuid
        yamlinfo['numcpus'] = config.hardware.numCPU
        yamlinfo['memory'] = config.hardware.memoryMB
        yamlinfo['status'] = translation[runtime.powerState]
        yamlinfo['nets'] = []
        yamlinfo['disks'] = []
        for entry in config.extraConfig:
            if entry.key in METADATA_FIELDS:
                yamlinfo[entry.key] = entry.value
            if entry.key == 'image':
                image = entry.value
                yamlinfo['user'] = common.get_user(entry.value)
            if entry.key == 'tags':
                yamlinfo['tags'] = entry.value
            if entry.key == 'ip':
                yamlinfo['ip'] = entry.value
        kubetype = yamlinfo.get('kubetype')
        cluster_network = yamlinfo.get('cluster_network')
        ips = []
        if runtime.powerState == "poweredOn":
            yamlinfo['host'] = runtime.host.name
            for nic in guest.net:
                if nic.ipAddress:
                    ip = nic.ipAddress[0]
                    if not ip.startswith('fe80::') and not ip.startswith('169.254') and not sdn_ip(ip, kubetype,
                                                                                                   cluster_network):
                        ips.append(ip)
                        if 'ip' not in yamlinfo:
                            yamlinfo['ip'] = ip
        if len(ips) > 1:
            yamlinfo['ips'] = ips
        if listinfo:
            return yamlinfo
        if image is None and kubetype is not None and kubetype == 'openshift':
            yamlinfo['user'] = 'core'
        if debug:
            yamlinfo['debug'] = config
        if not self.networks:
            self.set_networks()
        devices = config.hardware.device
        for number, dev in enumerate(devices):
            if "addressType" in dir(dev):
                try:
                    network = dev.backing.deviceName
                except:
                    switchuuid = dev.backing.port.switchUuid
                    portgroupkey = dev.backing.port.portgroupKey
                    for dvsnet in self.portgs:
                        if self.portgs[dvsnet][0] == switchuuid and self.portgs[dvsnet][1] == portgroupkey:
                            network = dvsnet
                device = dev.deviceInfo.label
                devicename = type(dev).__name__.replace('vim.vm.device.Virtual', '').lower()
                networktype = devicename
                mac = dev.macAddress
                net = {'device': device, 'mac': mac, 'net': network, 'type': networktype}
                yamlinfo['nets'].append(net)
            if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
                device = dev.deviceInfo.label
                disksize = convert(1000 * dev.capacityInKB, GB=False)
                diskformat = dev.backing.diskMode
                drivertype = 'thin' if dev.backing.thinProvisioned else 'thick'
                path = dev.backing.datastore.name
                disk = {'device': device, 'size': int(disksize), 'format': diskformat, 'type': drivertype,
                        'path': path}
                yamlinfo['disks'].append(disk)
            if isinstance(dev, vim.vm.device.VirtualCdrom)\
               and hasattr(dev.backing, 'fileName') and dev.backing.fileName is not None\
               and dev.backing.fileName.endswith('.iso'):
                yamlinfo['iso'] = dev.backing.fileName
        if obj.snapshot is not None and obj.snapshot.currentSnapshot is not None:
            yamlinfo['snapshot'] = obj.snapshot.rootSnapshotList[0].name
        return yamlinfo

    def list(self):
        si = self.si
        content = si.content
        vms = []
        vmFolder = self.basefolder
        all_vms = get_all_obj(content, [vim.VirtualMachine], folder=vmFolder)
        if not all_vms:
            return vms
        prop_collector = content.propertyCollector
        props = ['runtime', 'config', 'summary', 'guest']
        filter_spec = create_filter_spec(all_vms, props)
        options = vmodl.query.PropertyCollector.RetrieveOptions()
        vmlist = prop_collector.RetrievePropertiesEx([filter_spec], options)
        for o in vmlist.objects:
            obj = o.obj
            vmname = obj.name
            vm = convert_properties(o)
            summary, config = vm.get('summary'), vm.get('config')
            if summary is None or config is None:
                continue
            elif summary.runtime.connectionState != 'orphaned' and not config.template:
                if self.filtervms and 'plan' not in [x.key for x in config.extraConfig]:
                    continue
                try:
                    vms.append(self.info(vmname, vm=vm))
                except:
                    continue
        return sorted(vms, key=lambda x: x['name'])

    def list_pools(self):
        pools = []
        rootFolder = self.rootFolder
        si = self.si
        # dc = self.dc
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        for dts in clu.datastore:
            pools.append(dts.name)
            # datastorename = dts.name
            # total = dssize(dts)[0].replace('GB', '')
            # available = dssize(dts)[1].replace('GB', '')
            # results[datastorename] = [float(total), float(available), dc.name]
        return pools

    def beststorage(self):
        rootFolder = self.rootFolder
        si = self.si
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        bestds = ''
        bestsize = 0
        for dts in clu.datastore:
            datastorename = dts.name
            available = float(dssize(dts)[1].replace('GB', ''))
            if available > bestsize:
                bestsize = available
                bestds = datastorename
        return bestds

    def _getisos(self):
        rootFolder = self.rootFolder
        si = self.si
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        isos = []
        searchspec = vim.host.DatastoreBrowser.SearchSpec()
        filequery = [vim.host.DatastoreBrowser.IsoImageQuery(), vim.host.DatastoreBrowser.FolderQuery()]
        filequeryflags = vim.host.DatastoreBrowser.FileInfo.Details()
        filequeryflags.fileSize = True
        filequeryflags.modification = False
        filequeryflags.fileOwner = False
        filequeryflags.fileType = False
        searchspec.query = filequery
        searchspec.details = filequeryflags
        searchspec.sortFoldersFirst = True
        searchspec.searchCaseInsensitive = True
        for dts in clu.datastore:
            datastorename = dts.name
            datastorepath = f'[{datastorename}]'
            browser = dts.browser
            t = browser.SearchDatastore_Task(datastorepath, searchspec)
            waitForMe(t)
            result = t.info.result
            for element in result.file:
                folderpath = element.path
                if 'iso' in folderpath.lower():
                    t = browser.SearchDatastoreSubFolders_Task(f"{datastorepath}{folderpath}", searchspec)
                    waitForMe(t)
                    for r in t.info.result:
                        for isofile in r.file:
                            iso_path = f"{datastorepath}/{folderpath}/{isofile.path}"
                            if iso_path.endswith('.iso') and iso_path not in isos:
                                isos.append(iso_path)
        return isos

    def volumes(self, iso=False):
        if iso:
            return self._getisos()
        results = []
        si = self.si
        rootFolder = self.rootFolder
        o = si.content.viewManager.CreateContainerView(rootFolder, [vim.VirtualMachine], True)
        vmlist = o.view
        o.Destroy()
        vms = [v for v in vmlist if v.config.template and
               (v.summary is None or (v.summary is not None and v.summary.runtime.connectionState != 'orphaned'))]
        for v in sorted(vms, key=lambda x: x.name):
            devices = v.config.hardware.device
            for number, dev in enumerate(devices):
                if type(dev).__name__ == 'vim.vm.device.VirtualDisk':
                    prefix = '' if self.restricted else f'{dev.backing.datastore.name}/'
                    results.append(f'{prefix}{v.name}')
        return sorted(results)

    def update_metadata(self, name, metatype, metavalue, append=False):
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        configspec = vim.vm.ConfigSpec()
        opt = vim.option.OptionValue()
        opt.key = metatype
        opt.value = metavalue
        configspec.extraConfig = [opt]
        t = vm.ReconfigVM_Task(configspec)
        waitForMe(t)

    def update_memory(self, name, memory):
        if memory % 1024 != 0:
            warning("Rounding up memory to be multiple of 1024")
            memory += (1024 - memory) % 1024
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        configspec = vim.vm.ConfigSpec()
        configspec.memoryMB = memory
        t = vm.ReconfigVM_Task(configspec)
        waitForMe(t)

    def update_cpus(self, name, numcpus):
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        configspec = vim.vm.ConfigSpec()
        configspec.numCPUs = numcpus
        t = vm.ReconfigVM_Task(configspec)
        waitForMe(t)

    def update_start(self, name, start=True):
        print("not implemented")

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)
        return

    def update_iso(self, name, iso):
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if iso is not None:
            isos = [i for i in self._getisos() if i.endswith(iso)]
            if not isos:
                error(f"Iso {iso} not found.Leaving...")
                return {'result': 'failure', 'reason': f"Iso {iso} not found"}
            else:
                iso = isos[0]
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        c = changecd(self.si, vm, iso)
        if iso is None:
            answered = False
            timeout = 0
            while not answered:
                question = vm.runtime.question
                if question is not None:
                    pprint(f"Answering the following question automatically\n{vm.runtime.question.text}")
                    choice = vm.runtime.question.choice.choiceInfo[0].key
                    vm.AnswerVM(question.id, choice)
                    answered = True
                elif timeout > 20:
                    break
                else:
                    time.sleep(5)
        waitForMe(c)
        return {'result': 'success'}

    def convert_to_template(self, name):
        if self.esx:
            msg = "Operation not supported on single ESX"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        elif not self.esx:
            vm.MarkAsTemplate()

    def convert_to_vm(self, name):
        si = self.si
        dc = self.dc
        vmFolder = dc.vmFolder
        vm = findvm(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        else:
            rootFolder = self.rootFolder
            clu = find(si, rootFolder, vim.ComputeResource, self.clu)
            vm.MarkAsVirtualMachine(pool=clu.resourcePool)

    def dnsinfo(self, name):
        return None, None

    def _uploadimage(self, pool, origin, directory, isofolder=None):
        si = self.si
        rootFolder = self.rootFolder
        datastore = find(si, rootFolder, vim.Datastore, pool)
        if not datastore:
            msg = f"Pool {pool} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        destination = os.path.basename(origin)
        if isofolder is not None:
            directory = isofolder
        url = f"https://{self.vcip}:443/folder/{directory}/{destination}?dcPath={self.dc.name}&dsName={pool}"
        client_cookie = si._stub.cookie
        cookie_name = client_cookie.split("=", 1)[0]
        cookie_value = client_cookie.split("=", 1)[1].split(";", 1)[0]
        cookie_path = client_cookie.split("=", 1)[1].split(";", 1)[1].split(";", 1)[0].lstrip()
        cookie_text = " " + cookie_value + "; $" + cookie_path
        headers = {'Content-Type': 'application/octet-stream', 'Cookie': f"{cookie_name}={cookie_text}"}
        with open(origin, "rb") as f:
            context = ssl._create_unverified_context()
            try:
                req = urllib.request.Request(url, data=f, headers=headers, method='PUT')
                urllib.request.urlopen(req, context=context)
            except:
                url = url.replace('/folder', '')
                req = urllib.request.Request(url, data=f, headers=headers, method='PUT')
                try:
                    urllib.request.urlopen(req, context=context)
                except Exception as e:
                    error(f"Hit issue with with reason: {e}")

    def get_pool_path(self, pool):
        return pool

    def add_disk(self, name, size=1, pool=None, thin=True, image=None, shareable=False, existing=None,
                 interface='virtio', novm=False, overrides={}, diskname=None):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        config = info['config']
        spec = vim.vm.ConfigSpec()
        unit_number = 0
        for dev in config.hardware.device:
            if hasattr(dev.backing, 'fileName'):
                unit_number = int(dev.unitNumber) + 1
                if unit_number == 7:
                    unit_number = 8
            if isinstance(dev, vim.vm.device.VirtualSCSIController):
                controller = dev
        new_disk_kb = int(size) * 1024 * 1024
        disk_spec = vim.vm.device.VirtualDeviceSpec()
        disk_spec.fileOperation = "create"
        disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
        disk_spec.device = vim.vm.device.VirtualDisk()
        disk_spec.device.backing = vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
        disk_spec.device.backing.thinProvisioned = thin
        disk_spec.device.backing.diskMode = 'persistent'
        disk_spec.device.unitNumber = unit_number
        disk_spec.device.capacityInKB = new_disk_kb
        disk_spec.device.controllerKey = controller.key
        dev_changes = [disk_spec]
        spec.deviceChange = dev_changes
        t = vm.ReconfigVM_Task(spec=spec)
        waitForMe(t)
        return {'result': 'success'}

    def delete_disk(self, name=None, diskname=None, pool=None, novm=False):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        config = info['config']
        for dev in config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualDisk) and dev.deviceInfo.label == diskname:
                devspec = vim.vm.device.VirtualDeviceSpec()
                devspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
                devspec.device = dev
                spec = vim.vm.ConfigSpec()
                spec.deviceChange = [devspec]
                t = vm.ReconfigVM_Task(spec=spec)
                waitForMe(t)
                return {'result': 'success'}
        msg = f"Disk {diskname} not found in {name}"
        error(msg)
        return {'result': 'failure', 'reason': error}

    def detach_disks(self, name):
        pprint(f"Detaching non primary disks from {name}")
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        config = info['config']
        for index, dev in enumerate(config.hardware.device):
            if index > 0 and isinstance(dev, vim.vm.device.VirtualDisk):
                devspec = vim.vm.device.VirtualDeviceSpec()
                devspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
                devspec.device = dev
                devspec.fileOperation = None
                spec = vim.vm.ConfigSpec()
                spec.deviceChange = [devspec]
                t = vm.ReconfigVM_Task(spec=spec)
                waitForMe(t)
        return {'result': 'success'}

    def add_nic(self, name, network, model='virtio'):
        if network == 'default':
            network = 'VM Network'
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        config = info['config']
        spec = vim.vm.ConfigSpec()
        nicnumber = len([dev for dev in config.hardware.device if "addressType" in dir(dev)])
        nicname = 'Network adapter %d' % (nicnumber + 1)
        nicspec = createnicspec(nicname, network)
        nic_changes = [nicspec]
        spec.deviceChange = nic_changes
        t = vm.ReconfigVM_Task(spec=spec)
        waitForMe(t)
        return {'result': 'success'}

    def delete_nic(self, name, interface):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            msg = f"VM {name} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        config = info['config']
        for dev in config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualEthernetCard) and dev.deviceInfo.label == interface:
                devspec = vim.vm.device.VirtualDeviceSpec()
                devspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.remove
                devspec.device = dev
                spec = vim.vm.ConfigSpec()
                spec.deviceChange = [devspec]
                t = vm.ReconfigVM_Task(spec=spec)
                waitForMe(t)
                return {'result': 'success'}
        return {'result': 'failure', 'reason': f"Nic {interface} not found in {name}"}

    def list_networks(self):
        si = self.si
        rootFolder = si.content.rootFolder
        networks = {}
        view = si.content.viewManager.CreateContainerView(rootFolder, [vim.Network], True)
        netlist = collectproperties(si, view=view, objtype=vim.Network, pathset=['name'], includemors=True)
        for o in netlist:
            network = o['obj']
            cidr, dhcp, domainname = '', '', ''
            mode = 'accessible' if network.summary.accessible else 'notaccessible'
            networks[network.name] = {'cidr': cidr, 'dhcp': dhcp, 'domain': domainname, 'type': 'routed', 'mode': mode}
        if self.dvs:
            view = si.content.viewManager.CreateContainerView(rootFolder, [vim.dvs.DistributedVirtualPortgroup], True)
            dvslist = collectproperties(si, view=view, objtype=vim.dvs.DistributedVirtualPortgroup, pathset=['name'],
                                        includemors=True)
            for o in dvslist:
                network = o['obj']
                cidr, dhcp, domainname, mode = '', '', '', ''
                networks[network.name] = {'cidr': cidr, 'dhcp': dhcp, 'domain': domainname, 'type': 'routed',
                                          'mode': mode}
        return networks

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        return networkinfo

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        si = self.si
        cluster = self.clu
        networkFolder = self.dc.networkFolder
        rootFolder = self.rootFolder
        net = find(si, rootFolder, vim.Network, name)
        if net is not None:
            msg = f"Network {name} already there"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        o = si.content.viewManager.CreateContainerView(rootFolder, [vim.DistributedVirtualSwitch], True)
        dvnetworks = o.view
        o.Destroy()
        for dvnetw in dvnetworks:
            for portg in dvnetw.portgroup:
                if portg.name == name:
                    msg = f"Network {name} already there"
                    error(msg)
                    return {'result': 'failure', 'reason': msg}
        if overrides.get('distributed', False):
            pnic_specs = []
            dvs_host_configs = []
            uplink_port_names = []
            dvs_create_spec = vim.DistributedVirtualSwitch.CreateSpec()
            dvs_config_spec = vim.DistributedVirtualSwitch.ConfigSpec()
            dvs_config_spec.name = name
            dvs_config_spec.uplinkPortPolicy = vim.DistributedVirtualSwitch.NameArrayUplinkPortPolicy()
            for x in range(len(cluster.host)):
                uplink_port_names.append("dvUplink%d" % x)
            for host in cluster.host:
                dvs_config_spec.uplinkPortPolicy.uplinkPortName = uplink_port_names
                dvs_config_spec.maxPorts = 2000
                pnic_spec = vim.dvs.HostMember.PnicSpec()
                pnic_spec.pnicDevice = 'vmnic1'
                pnic_specs.append(pnic_spec)
                dvs_host_config = vim.dvs.HostMember.ConfigSpec()
                dvs_host_config.operation = vim.ConfigSpecOperation.add
                dvs_host_config.host = host
                dvs_host_configs.append(dvs_host_config)
                dvs_host_config.backing = vim.dvs.HostMember.PnicBacking()
                dvs_host_config.backing.pnicSpec = pnic_specs
                dvs_config_spec.host = dvs_host_configs
                dvs_create_spec.configSpec = dvs_config_spec
            dvs_create_spec.productInfo = vim.dvs.ProductSpec(version='5.1.0')
            networkFolder.CreateDistributedVirtualSwitch()
        else:
            return {'result': 'failure', 'reason': "Not implemented yet for non dvs networks"}
        return {'result': 'success'}

    def delete_network(self, name=None, cidr=None, force=False):
        si = self.si
        rootFolder = self.rootFolder
        try:
            net = find(si, rootFolder, vim.dvs.DistributedVirtualPortgroup, name)
            net.Destroy()
        except:
            try:
                net = find(si, rootFolder, vim.Network, name)
                net.Destroy()
            except:
                msg = f"Network {name} not found"
                error(msg)
                return {'result': 'failure', 'reason': msg}
        return {'result': 'success'}

    def vm_ports(self, name):
        return ['default']

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        si = self.si
        rootFolder = self.rootFolder
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        resourcepool = clu.resourcePool
        vmFolder = self.basefolder
        manager = si.content.ovfManager
        shortimage = os.path.basename(url).split('?')[0]
        name = name.replace('.ova', '').replace('.x86_64', '') if name is not None else shortimage
        iso = True if shortimage.endswith('.iso') or name.endswith('.iso') else False
        if not shortimage.endswith('ova') and not shortimage.endswith('zip') and not iso\
           and which('qemu-img') is None:
            msg = "qemu-img is required for conversion"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        if shortimage in [os.path.basename(v) for v in self.volumes()]:
            pprint(f"Template {shortimage} already there")
            return {'result': 'success'}
        if not find(si, rootFolder, vim.Datastore, pool):
            msg = f"Pool {pool} not found"
            error(msg)
            return {'result': 'failure', 'reason': msg}
        if os.path.exists(url):
            pprint(f"Using {url} as path")
        elif not os.path.exists(f'/tmp/{shortimage}'):
            pprint(f"Downloading locally {shortimage}")
            downloadcmd = f"curl -kLo /tmp/{shortimage} -f '{url}'"
            code = os.system(downloadcmd)
            if code != 0:
                msg = "Unable to download indicated image"
                error(msg)
                return {'result': 'failure', 'reason': msg}
        else:
            pprint(f"Using found /tmp/{shortimage}")
        if iso:
            isofile = os.path.abspath(url) if os.path.exists(url) else f"/tmp/{shortimage}"
            if name is not None:
                new_file = f"{os.path.dirname(isofile)}/{name}"
                os.rename(isofile, new_file)
                isofile = new_file
            if self.isofolder is not None:
                isofolder = self.isofolder.split('/')
                isopool = re.sub(r"[\[\]]", '', isofolder[0])
                isofolder = isofolder[1]
            else:
                isopool = pool
                isofolder = None
            destination = '' if isofolder is not None else name
            self._uploadimage(isopool, isofile, destination, isofolder=isofolder)
            return {'result': 'success'}
        vmdk_path = None
        ovf_path = None
        basedir = os.path.dirname(os.path.abspath(url)) if os.path.exists(url) else '/tmp'
        if url.endswith('zip'):
            with ZipFile(f"{basedir}/{shortimage}") as zipf:
                for _fil in zipf.namelist():
                    if _fil.endswith('vmdk'):
                        vmdk_path = f'{basedir}/{_fil}'
                    elif _fil.endswith('ovf'):
                        ovf_path = f'{basedir}/{_fil}'
                if vmdk_path is None or ovf_path is None:
                    msg = "Incorrect ova file"
                    error(msg)
                    return {'result': 'failure', 'reason': msg}
                zipf.extractall(basedir)
        elif url.endswith('ova'):
            with tarfile.open(f"{basedir}/{shortimage}") as tar:
                for _fil in [x.name for x in tar.getmembers()]:
                    if _fil.endswith('vmdk'):
                        vmdk_path = f'{basedir}/{_fil}'
                    elif _fil.endswith('ovf'):
                        ovf_path = f'{basedir}/{_fil}'
                if vmdk_path is None or ovf_path is None:
                    msg = "Incorrect ova file"
                    error(msg)
                    return {'result': 'failure', 'reason': msg}
                tar.extractall(basedir)
        else:
            need_uncompress = any(shortimage.endswith(suffix) for suffix in ['.gz', '.xz', '.bz2', '.zst'])
            if need_uncompress:
                extension = os.path.splitext(shortimage)[1].replace('.', '')
                executable = {'xz': 'unxz', 'gz': 'gunzip', 'bz2': 'bunzip2', 'zst': 'zstd'}
                flag = '--decompress' if extension == 'zstd' else '-f'
                executable = executable[extension]
                uncompresscmd = f"{executable} {flag} {basedir}/{shortimage}"
                os.system(uncompresscmd)
                shortimage = shortimage.replace(f'.{extension}', '')
            if '.' in shortimage:
                extension = os.path.splitext(shortimage)[1].replace('.', '')
                vmdk_file = shortimage.replace(extension, 'vmdk')
            else:
                vmdk_file = f"{shortimage}.vmdk"
            vmdk_path = f"{basedir}/{vmdk_file}"
            if cmds and shortimage.endswith('qcow2') and which('virt-customize') is not None:
                for cmd in cmds:
                    cmd = f"virt-customize -a {basedir}/{shortimage} --run-command '{cmd}'"
                    os.system(cmd)
            if not os.path.exists(vmdk_path):
                pprint("Converting qcow2 file to vmdk")
                cmd = f"qemu-img convert -O vmdk -o subformat=streamOptimized {basedir}/{shortimage} {vmdk_path}"
                os.popen(cmd).read()
            ovf_path = vmdk_path.replace('.vmdk', '.ovf')
            commondir = os.path.dirname(common.pprint.__code__.co_filename)
            vmdk_info = json.loads(os.popen(f"qemu-img info {vmdk_path} --output json").read())
            virtual_size = vmdk_info['virtual-size']
            actual_size = vmdk_info['actual-size']
            ovfcontent = open(f"{commondir}/vm.ovf.j2").read().format(name=shortimage, virtual_size=virtual_size,
                                                                      actual_size=actual_size, vmdk_file=vmdk_file,
                                                                      import_network=self.import_network)
            with open(ovf_path, 'w') as f:
                f.write(ovfcontent)
        ovfd = open(ovf_path).read()
        ovfd = re.sub('<Name>.*</Name>', f'<Name>{name}</Name>', ovfd)
        datastore = find(si, rootFolder, vim.Datastore, pool)
        network = None
        for host in self._get_hosts(self.clu):
            networks = [n for n in host.network if n.name == self.import_network]
            if networks:
                pprint(f"Using esxi host {host.name} for import")
                network = networks[0]
                break
        if network is None:
            error(f"Couldn't find any esxi host in the cluster with network {self.import_network}")
            error("Set import_network to a valid value")
            sys.exit(1)
        networkmapping = vim.OvfManager.NetworkMapping.Array()
        nm = vim.OvfManager.NetworkMapping(name="VM Network", network=network)
        networkmapping.append(nm)
        spec_params = vim.OvfManager.CreateImportSpecParams(diskProvisioning="thin", networkMapping=networkmapping,
                                                            hostSystem=host)
        import_spec = manager.CreateImportSpec(ovfd, resourcepool, datastore, spec_params)
        if import_spec.error:
            error(f"Import spec error: {import_spec.error}")
            sys.exit(1)
        lease = resourcepool.ImportVApp(import_spec.importSpec, vmFolder)
        while True:
            if lease.state == vim.HttpNfcLease.State.ready:
                url = lease.info.deviceUrl[0].url.replace('*', host.name)
                pprint(f"Uploading {vmdk_path} to {url}")
                keepalive_thread = Thread(target=keep_lease_alive, args=(lease,))
                keepalive_thread.start()
                cmd = f'curl -k -# -X POST -T {vmdk_path} -H "Content-Type: application/x-vnd.vmware-streamVmdk" {url}'
                call(cmd, shell=True)
                lease.HttpNfcLeaseComplete()
                keepalive_thread.join()
                break
            elif lease.state == vim.HttpNfcLease.State.error:
                error(f"Lease error: {lease.error}")
                sys.exit(1)
        if not self.esx:
            self.convert_to_template(name)
            if os.path.exists(f'{basedir}/{shortimage}'):
                os.remove(f'{basedir}/{shortimage}')
            if os.path.exists(vmdk_path):
                os.remove(vmdk_path)
        return {'result': 'success'}

    def _get_hosts(self, cluster):
        si = self.si
        rootFolder = self.rootFolder
        o = si.content.viewManager.CreateContainerView(rootFolder, [vim.ComputeResource], True)
        view = o.view
        o.Destroy()
        for clu in view:
            if clu.name == cluster:
                return clu.host
        return []

    def info_host(self):
        data = {}
        si = self.si
        about = si.content.about
        data['vcenter'] = self.vcip
        data['version'] = about.version
        data['api_version'] = about.apiVersion
        rootFolder = self.rootFolder
        o = si.content.viewManager.CreateContainerView(rootFolder, [vim.Datacenter], True)
        view = o.view
        o.Destroy()
        datacenters = []
        for datacenter in view:
            new_datacenter = {}
            new_datacenter['datacenter'] = datacenter.name
            clusters = []
            for clu in datacenter.hostFolder.childEntity:
                new_clu = {}
                new_clu['cluster'] = clu.name
                datastores = []
                for dts in clu.datastore:
                    datastores.append(dts.name)
                new_clu['datastores'] = datastores
                hosts = []
                for h in clu.host:
                    hosts.append(h.name)
                new_clu['datastores'] = datastores
                networks = []
                for n in clu.network:
                    networks.append(n.name)
                new_clu['networks'] = networks
                clusters.append(new_clu)
            new_datacenter['clusters'] = clusters
            datacenters.append(new_datacenter)
        data['datacenters'] = datacenters
        return data

    def delete_image(self, image, pool=None):
        si = self.si
        if image.endswith('.iso'):
            matching_isos = [iso for iso in self._getisos() if iso.endswith(image)]
            if not matching_isos:
                return {'result': 'failure', 'reason': f'Iso {image} not found'}
            else:
                isopath = matching_isos[0]
                deletedirectory(si, self.dc, isopath)
                return {'result': 'success'}
        if '/' in image:
            pool, image = image.split('/')
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, image)
        if vm is None or not info['config'].template:
            return {'result': 'failure', 'reason': f'Image {image} not found'}
        else:
            t = vm.Destroy_Task()
            waitForMe(t)
            return {'result': 'success'}

    def export(self, name, image=None):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, name)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        if info['runtime'].powerState == "poweredOn":
            t = vm.PowerOffVM_Task()
            waitForMe(t)
        vm.MarkAsTemplate()
        if image is not None:
            vm.Rename(image)
        return {'result': 'success'}

    def list_dns(self, domain):
        return []

    def create_bucket(self, bucket, public=False):
        print("not implemented")

    def delete_bucket(self, bucket):
        print("not implemented")

    def delete_from_bucket(self, bucket, path):
        print("not implemented")

    def download_from_bucket(self, bucket, path):
        print("not implemented")

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        print("not implemented")

    def list_buckets(self):
        print("not implemented")
        return []

    def list_bucketfiles(self, bucket):
        print("not implemented")
        return []

    def get_guestid(image):
        if 'centos' in image.lower():
            guestid = 'centos8_64Guest'
        elif 'rhel' in image.lower() or 'rhcos' in image.lower():
            guestid = 'rhel8_64Guest'
        elif 'fedora' in image.lower() or 'fos' in image.lower():
            guestid = 'fedora64Guest'
        elif 'ubuntu' in image.lower():
            guestid = 'ubuntu64Guest'
        elif 'debian' in image.lower():
            guestid = 'debian10_64Guest'
        else:
            guestid = 'genericLinuxGuest'
        return guestid

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False):
        print("not implemented")

    def update_nic(self, name, index, network):
        print("not implemented")

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def clone(self, old, new, full=False, start=False):
        si = self.si
        restricted = self.restricted
        rootFolder = self.rootFolder
        vmFolder = self.basefolder
        old_info = self.info(old)
        plan, cluster = old_info.get('plan'), old_info.get('cluster')
        if not restricted:
            if cluster is not None:
                createfolder(si, vmFolder, cluster)
                vmFolder = find(si, vmFolder, vim.Folder, cluster)
            elif plan != 'kvirt':
                createfolder(si, vmFolder, plan)
                vmFolder = find(si, vmFolder, vim.Folder, plan)
        si = self.si
        clu = find(si, rootFolder, vim.ComputeResource, self.clu)
        resourcepool = clu.resourcePool
        imageobj = findvm(si, rootFolder, old)
        if imageobj is None:
            return {'result': 'failure', 'reason': f"VM {old} not found"}
        clonespec = createclonespec(resourcepool)
        confspec = vim.vm.ConfigSpec()
        confspec.annotation = new
        extraconfig = []
        clonespec.powerOn = start
        confspec.extraConfig = extraconfig
        t = imageobj.CloneVM_Task(folder=vmFolder, name=new, spec=clonespec)
        waitForMe(t)
        return {'result': 'success'}

    def create_snapshot(self, name, base):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, base)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {base} not found"}
        description = f"Snapshot {name}"
        dump_memory = False
        quiesce = False
        t = vm.CreateSnapshot(name, description, dump_memory, quiesce)
        waitForMe(t)
        return {'result': 'success'}

    def delete_snapshot(self, name, base):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, base)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {base} not found"}
        snapshots = vm.snapshot.rootSnapshotList if vm.snapshot is not None else []
        for snapshot in snapshots:
            if snapshot.name == name:
                t = snapshot.snapshot.RemoveSnapshot_Task(True)
                waitForMe(t)
                return {'result': 'success'}
        return {'result': 'failure', 'reason': f'Snapshot {name} not found'}

    def list_snapshots(self, base):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, base)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {base} not found"}
        snapshots = vm.snapshot.rootSnapshotList if vm.snapshot is not None else []
        return [snapshot.name for snapshot in snapshots]

    def revert_snapshot(self, name, base):
        si = self.si
        vmFolder = self.basefolder
        vm, info = findvm2(si, vmFolder, base)
        if vm is None:
            return {'result': 'failure', 'reason': f"VM {base} not found"}
        snapshots = vm.snapshot.rootSnapshotList if vm.snapshot is not None else []
        for snapshot in snapshots:
            if snapshot.name == name:
                t = snapshot.snapshot.RevertToSnapshot_Task()
                waitForMe(t)
                return {'result': 'success'}
        return {'result': 'failure', 'reason': f'Snapshot {name} not found'}

    def ip(self, name):
        result = None
        si = self.si
        vmFolder = self.basefolder
        obj, vm = findvm2(si, vmFolder, name)
        guest = vm['guest']
        if vm is not None:
            runtime = vm['runtime']
            guest = vm['guest']
            if runtime.powerState == "poweredOn":
                for nic in guest.net:
                    if nic.ipAddress:
                        result = nic.ipAddress[0]
                        break
        return result

    def create_vm_folder(self, name):
        si = self.si
        vmFolder = self.basefolder
        if find(si, vmFolder, vim.Folder, name) is None:
            createfolder(si, vmFolder, name)

    def list_flavors(self):
        return []

    def list_security_groups(self, network=None):
        print("not implemented")
        return []

    def create_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_security_group(self, name):
        print("not implemented")
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_disks(self):
        print("not implemented")

    def list_subnets(self):
        print("not implemented")
        return {}

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        print("not implemented")

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")

    def delete_pool(self, name, full=False):
        print("not implemented")

    def disk_exists(self, pool, name):
        print("not implemented")

    def network_ports(self, name):
        print("not implemented")

    def update_flavor(self, name, flavor):
        print("not implemented")
        return {'result': 'success'}

    def _datastores_datacenters(self):
        si = self.si
        rootFolder = self.rootFolder
        o = si.content.viewManager.CreateContainerView(rootFolder, [vim.Datacenter], True)
        view = o.view
        o.Destroy()
        datastores = {}
        for datacenter in view:
            for clu in datacenter.hostFolder.childEntity:
                if 'ClusterComputeResource' not in str(clu.__class__):
                    continue
                for datastore in clu.datastore:
                    datastores[datastore.name] = datacenter.name
        return datastores

    def info_subnet(self, name):
        print("not implemented")
        return {}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        print("not implemented")
        return []

    def set_macs(self, name, macs):
        vm = find(self.si, self.dc.vmFolder, vim.VirtualMachine, name)
        currentdevices = vm.config.hardware.device
        currentnics = [d for d in currentdevices if isinstance(d, vim.vm.device.VirtualEthernetCard)]
        confspec = vim.vm.ConfigSpec()
        devconfspec = []
        for index in macs:
            currentnic = currentnics[index]
            currentnic.macAddress = macs[index]
            currentnic.addressType = vim.vm.device.VirtualEthernetCardOption.MacTypes.manual
            nicspec = vim.vm.device.VirtualDeviceSpec(device=currentnic, operation="edit")
            devconfspec.append(nicspec)
        confspec.deviceChange = devconfspec
        t = vm.Reconfigure(confspec)
        waitForMe(t)

    def reconnect_hosts(self):
        si = self.si
        rootFolder = self.rootFolder
        view = si.content.viewManager.CreateContainerView(rootFolder, [vim.HostSystem], True)
        for host in view.view:
            pprint(f"Reconnecting Host {host.name}")
            host.Reconnect()



from ipaddress import ip_network
from kvirt.common import error, success, pprint, warning, get_kubectl, info2, container_mode, create_app_generic
from kvirt.common import deploy_cloud_storage, wait_cloud_dns, update_etc_hosts, fix_typos, get_new_vip
from kvirt.common import wait_for_nodes
import os
import re
from random import choice
from shutil import which
from string import ascii_letters, digits
from subprocess import call, run
from tempfile import NamedTemporaryFile
from time import sleep
from yaml import safe_dump, safe_load

cloud_providers = ['aws', 'azure', 'gcp', 'ibm', 'hcloud']


def update_ip_alias(config, nodes):
    failure = False
    timeout = 0
    cmd_one = ['kubectl', 'get', 'nodes', '-o=jsonpath={range .items[?(@.spec.podCIDR)]}{.metadata.name}{\'\\n\'}{end}']
    while True:
        if timeout > 240:
            error(f"Timeout waiting for {nodes} nodes to have a Pod cidr assigned")
            return 1
        pprint(f"Waiting 5s for {nodes} nodes to have a Pod cidr assigned")
        result = run(cmd_one, capture_output=True, text=True)
        current_nodes = len(result.stdout.splitlines())
        if current_nodes == nodes:
            break
        else:
            sleep(5)
            timeout += 5
    for node in safe_load(os.popen("kubectl get node -o yaml").read())['items']:
        name = node['metadata']['name']
        try:
            pod_cidr = node['spec']['podCIDR']
            config.k.update_aliases(name, pod_cidr)
        except KeyError as e:
            error(f"Hit Error with node {name}: {e}")
            failure = True
    return 1 if failure else 0


def scale(config, plandir, cluster, overrides):
    storedparameters = overrides.get('storedparameters', True)
    provider = config.type
    plan = cluster
    data = {'cluster': cluster, 'domain': 'karmalabs.corp', 'image': 'ubuntu2004', 'kube': cluster, 'kubetype': 'k3s',
            'sdn': 'flannel', 'extra_scripts': [], 'cloud_native': False, 'ctlplanes': 1, 'workers': 0}
    data['basedir'] = '/workdir' if container_mode() else '.'
    cluster = data['cluster']
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if storedparameters and os.path.exists(f"{clusterdir}/kcli_parameters.yml"):
        with open(f"{clusterdir}/kcli_parameters.yml", 'r') as install:
            installparam = safe_load(install)
            data.update(installparam)
            plan = installparam.get('plan', plan)
    else:
        data['client'] = config.client
    data.update(overrides)
    data['scale'] = True
    cloud_native = data.get('cloud_native')
    cloud_lb = data.get('cloud_lb', provider in cloud_providers and data['ctlplanes'] > 1)
    data['cloud_lb'] = cloud_lb
    ctlplanes = data['ctlplanes']
    workers = data['workers']
    sdn = None if 'sdn' in overrides and overrides['sdn'] is None else data.get('sdn')
    client = config.client
    if 'api_ip' not in data:
        data['api_ip'] = None
    if 'first_ip' not in data or data['first_ip'] is None:
        first_info = config.k.info(f'{cluster}-ctlplane-0') or config.k.info(f'{cluster}-master-0')
        data['first_ip'] = first_info.get('private_ip') or first_info.get('ip')
    pprint(f"Scaling on client {client}")
    if os.path.exists(clusterdir):
        with open(f"{clusterdir}/kcli_parameters.yml", 'w') as paramfile:
            safe_dump(data, paramfile)
    vmrules_all_names = []
    if data.get('vmrules', config.vmrules) and data.get('vmrules_strict', config.vmrules_strict):
        vmrules_all_names = [list(entry.keys())[0] for entry in data.get('vmrules', config.vmrules)]
    for role in ['ctlplanes', 'workers']:
        install_k3s_args = []
        for arg in data:
            if arg.startswith('install_k3s'):
                install_k3s_args.append(f"{arg.upper()}={data[arg]}")
        overrides = data.copy()
        threaded = data.get('threaded', False) or data.get(f'{role}_threaded', False)
        if role == 'ctlplanes':
            if ctlplanes == 1:
                continue
            if provider not in cloud_providers and not cloud_lb\
               and ('virtual_router_id' not in overrides or 'auth_pass' not in overrides):
                warning("Scaling up of ctlplanes won't work without virtual_router_id and auth_pass")
            if sdn is None or sdn != 'flannel':
                install_k3s_args.append("INSTALL_K3S_EXEC='--flannel-backend=none'")
            install_k3s_args = ' '.join(install_k3s_args)
        if role == 'workers' and workers == 0:
            continue
        if vmrules_all_names:
            reg = re.compile(f'{cluster}-{role[:-1]}-[0-9]+')
            vmrules_names = [name for name in vmrules_all_names if reg.match(name)]
            if len(vmrules_names) != overrides.get(role, 1):
                warning(f"Adjusting {role} number to vmrule entries")
                overrides[role] = len(vmrules_names)
            overrides['vmrules_names'] = sorted(vmrules_names)
        overrides['install_k3s_args'] = install_k3s_args
        result = config.plan(plan, inputfile=f'{plandir}/{role}.yml', overrides=overrides, threaded=threaded)
        if result['result'] != 'success':
            return result
        elif result.get('newvms', []):
            pprint(f"{role.capitalize()} nodes will join the cluster in a few minutes")
    if cloud_native and provider == 'gcp':
        pprint("Updating ip alias ranges")
        result = update_ip_alias(config, ctlplanes + workers)
        if result != 0:
            return {'result': 'failure', 'reason': "Issue when updating ip alias ranges"}
    return {'result': 'success'}


def create(config, plandir, cluster, overrides):
    provider = config.type
    k = config.k
    data = safe_load(open(f'{plandir}/kcli_default.yml'))
    data.update(overrides)
    fix_typos(data)
    storedparameters = data.get('storedparameters', True)
    cloud_dns = data['cloud_dns']
    data['cloud_lb'] = overrides.get('cloud_lb', provider in cloud_providers and data['ctlplanes'] > 1)
    cloud_lb = data['cloud_lb']
    cloud_storage = data['cloud_storage']
    cloud_native = data['cloud_native']
    data['cluster'] = cluster
    plan = cluster
    data['kube'] = data['cluster']
    data['kubetype'] = 'k3s'
    autoscale = data['autoscale']
    ctlplanes = data['ctlplanes']
    workers = data['workers']
    network = data['network']
    sdn = None if 'sdn' in overrides and overrides['sdn'] is None else data.get('sdn')
    domain = data['domain']
    image = data['image']
    api_ip = data.get('api_ip')
    if ctlplanes > 1:
        if provider in cloud_providers:
            if not cloud_lb:
                msg = "Multiple ctlplanes require cloud_lb to be set to True"
                return {'result': 'failure', 'reason': msg}
            elif api_ip is None:
                api_ip = f"api.{cluster}.{domain}"
                data['api_ip'] = api_ip
        elif api_ip is None:
            networkinfo = k.info_network(network)
            if provider == 'kvm' and networkinfo['type'] == 'routed':
                cidr = networkinfo['cidr']
                if cidr == 'N/A':
                    msg = "Couldnt gather an api_ip from your specified network"
                    return {'result': 'failure', 'reason': msg}
                api_index = 2 if ':' in cidr else -3
                api_ip = str(get_new_vip(network, ipv6=':' in cidr) or ip_network(cidr)[api_index])
                warning(f"Using {api_ip} as api_ip")
                data['api_ip'] = api_ip
                data['automatic_api_ip'] = True
            elif provider == 'kubevirt':
                selector = {'kcli/plan': plan, 'kcli/role': 'ctlplane'}
                api_ip = config.k.create_service(f"{cluster}-api", config.k.namespace, selector,
                                                 _type="LoadBalancer", ports=[6443])
                if api_ip is None:
                    msg = "Couldnt get an kubevirt api_ip from service"
                    return {'result': 'failure', 'reason': msg}
                else:
                    pprint(f"Using api_ip {api_ip}")
                    data['api_ip'] = api_ip
            else:
                msg = "You need to define api_ip in your parameters file"
                return {'result': 'failure', 'reason': msg}
        if not cloud_lb and ctlplanes > 1:
            if data.get('virtual_router_id') is None:
                data['virtual_router_id'] = hash(data['cluster']) % 254 + 1
                pprint(f"Using keepalived virtual_router_id {data['virtual_router_id']}")
            if data.get('auth_pass') is None:
                auth_pass = ''.join(choice(ascii_letters + digits) for i in range(5))
                data['auth_pass'] = auth_pass
    install_k3s_args = []
    for arg in data:
        if arg.startswith('install_k3s'):
            install_k3s_args.append(f"{arg.upper()}={data[arg]}")
    cluster = data.get('cluster')
    clusterdir = os.path.expanduser(f"~/.kcli/clusters/{cluster}")
    if os.path.exists(clusterdir):
        msg = f"Remove existing directory {clusterdir} or use --force"
        return {'result': 'failure', 'reason': msg}
    if which('kubectl') is None:
        get_kubectl()
    if not os.path.exists(clusterdir):
        os.makedirs(clusterdir)
        os.mkdir(f"{clusterdir}/auth")
    for arg in data.get('extra_ctlplane_args', []):
        if arg.startswith('--data-dir='):
            data['data_dir'] = arg.split('=')[1]
    bootstrap_overrides = data.copy()
    if os.path.exists("manifests") and os.path.isdir("manifests"):
        bootstrap_overrides['files'] = [{"path": "/root/manifests", "currentdir": True, "origin": "manifests"}]
    bootstrap_install_k3s_args = install_k3s_args.copy()
    if sdn is None or sdn != 'flannel':
        bootstrap_install_k3s_args.append("INSTALL_K3S_EXEC='--flannel-backend=none'")
    bootstrap_install_k3s_args = ' '.join(bootstrap_install_k3s_args)
    bootstrap_overrides['install_k3s_args'] = bootstrap_install_k3s_args
    result = config.plan(plan, inputfile=f'{plandir}/bootstrap.yml', overrides=bootstrap_overrides)
    if result['result'] != "success":
        return result
    first_info = config.k.info(f'{cluster}-ctlplane-0')
    first_ip = first_info.get('private_ip') or first_info.get('ip')
    data['first_ip'] = first_ip
    if storedparameters:
        with open(f"{clusterdir}/kcli_parameters.yml", 'w') as p:
            installparam = overrides.copy()
            installparam['client'] = config.client
            installparam['cluster'] = cluster
            installparam['api_ip'] = api_ip
            installparam['first_ip'] = first_ip
            installparam['plan'] = plan
            installparam['kubetype'] = 'k3s'
            installparam['image'] = image
            if not cloud_lb and ctlplanes > 1:
                installparam['virtual_router_id'] = data['virtual_router_id']
                installparam['auth_pass'] = data['auth_pass']
            if 'automatic_api_ip' in data:
                installparam['automatic_api_ip'] = True
            safe_dump(installparam, p, default_flow_style=False, encoding='utf-8', allow_unicode=True)
    for role in ['ctlplanes', 'workers']:
        if (role == 'ctlplanes' and ctlplanes == 1) or (role == 'workers' and workers == 0):
            continue
        nodes_overrides = data.copy()
        nodes_install_k3s_args = install_k3s_args.copy()
        nodes_overrides['install_k3s_args'] = nodes_install_k3s_args
        if role == 'ctlplanes':
            if sdn is None or sdn != 'flannel':
                nodes_install_k3s_args.append("INSTALL_K3S_EXEC='--flannel-backend=none'")
            nodes_install_k3s_args = ' '.join(nodes_install_k3s_args)
            nodes_overrides['install_k3s_args'] = nodes_install_k3s_args
            pprint("Deploying extra ctlplanes")
            threaded = data.get('threaded', False) or data.get('ctlplanes_threaded', False)
            config.plan(plan, inputfile=f'{plandir}/ctlplanes.yml', overrides=nodes_overrides, threaded=threaded)
        if role == 'workers':
            pprint("Deploying workers")
            os.chdir(os.path.expanduser("~/.kcli"))
            threaded = data.get('threaded', False) or data.get('workers_threaded', False)
            config.plan(plan, inputfile=f'{plandir}/workers.yml', overrides=nodes_overrides, threaded=threaded)
    if cloud_lb and provider in cloud_providers:
        if cloud_dns:
            config.k.delete_dns(f'api.{cluster}', domain=domain)
        if provider == 'aws':
            data['vpcid'] = config.k.get_vpcid_of_vm(f"{cluster}-ctlplane-0")
        result = config.plan(plan, inputfile=f'{plandir}/cloud_lb_api.yml', overrides=data)
        if result['result'] != 'success':
            return result
    success(f"K3s cluster {cluster} deployed!!!")
    info2(f"export KUBECONFIG=$HOME/.kcli/clusters/{cluster}/auth/kubeconfig")
    info2("export PATH=$PWD:$PATH")
    if provider in cloud_providers and cloud_lb:
        if cloud_dns:
            wait_cloud_dns(cluster, domain)
        elif api_ip is not None and api_ip == f'api.{cluster}.{domain}':
            for lbentry in config.list_loadbalancers():
                if lbentry[0] == f'api-{cluster}':
                    lb_ip = lbentry[1]
                    update_etc_hosts(cluster, domain, lb_ip)
                    break
    os.environ['KUBECONFIG'] = f"{clusterdir}/auth/kubeconfig"
    apps = data.get('apps', [])
    if apps:
        appdir = f"{plandir}/apps"
        os.environ["PATH"] = f'{os.getcwd()}:{os.environ["PATH"]}'
        for app in apps:
            app_data = data.copy()
            if not os.path.exists(appdir):
                warning(f"Skipping unsupported app {app}")
            else:
                pprint(f"Adding app {app}")
                if f'{app}_version' not in overrides:
                    app_data[f'{app}_version'] = 'latest'
                create_app_generic(config, app, appdir, overrides=app_data)
    if ctlplanes + workers > 1:
        ready = wait_for_nodes(ctlplanes + workers)
        if not ready:
            msg = "Timeout waiting for all nodes to join"
            return {'result': 'failure', 'reason': msg}
    if autoscale:
        config.import_in_kube(network=network, secure=True)
        with NamedTemporaryFile(mode='w+t') as temp:
            commondir = os.path.dirname(pprint.__code__.co_filename)
            autoscale_overrides = {'cluster': cluster, 'kubetype': 'k3s', 'workers': workers, 'replicas': 1}
            autoscale_data = config.process_inputfile(cluster, f"{commondir}/autoscale.yaml.j2",
                                                      overrides=autoscale_overrides)
            temp.write(autoscale_data)
            autoscalecmd = f"kubectl create -f {temp.name}"
            call(autoscalecmd, shell=True)
    if provider in cloud_providers:
        if cloud_storage and provider == 'aws':
            pprint("Deploying cloud storage class")
            deploy_cloud_storage(config, cluster)
        if cloud_native and provider == 'gcp':
            pprint("Updating ip alias ranges")
            result = update_ip_alias(config, ctlplanes + workers)
            if result != 0:
                return {'result': 'failure', 'reason': "Issue when updating ip alias ranges"}
    return {'result': 'success'}



from base64 import b64encode
import json
from kvirt.common import error, pprint
from urllib.request import urlopen, Request
import ssl


class KsphereTag:
    def __init__(self, vcip, user, password):
        ssl._create_default_https_context = ssl._create_unverified_context
        self.url = f'https://{vcip}/rest/com/vmware/cis/tagging'
        auth_url = f'https://{vcip}/rest/com/vmware/cis/session'
        b64_auth = b64encode(f"{user}:{password}".encode()).decode()
        headers = {'Authorization': f"Basic {b64_auth}"}
        request = Request(auth_url, headers=headers, method='POST')
        resp = urlopen(request)
        status_code = resp.getcode()
        if status_code != 200:
            error(f'Error! API responded with: {status_code}')
            error(resp.reason)
            return
        self.sid = json.loads(resp.read())['value']

    def _delete(self, url):
        request = Request(url, headers={'vmware-api-session-id': self.sid}, method='DELETE')
        resp = urlopen(request)
        if resp.getcode() not in [200, 201]:
            error(resp.reason)

    def _get(self, url):
        request = Request(url, headers={'vmware-api-session-id': self.sid})
        resp = urlopen(request)
        if resp.getcode() not in [200, 201]:
            error(resp.reason)
            return None
        else:
            return json.loads(resp.read())

    def _post(self, url, data):
        data = json.dumps(data).encode('utf-8')
        request = Request(url, data=data, headers={'vmware-api-session-id': self.sid,
                                                   'content-type': 'application/json'})
        resp = urlopen(request)
        if resp.getcode() not in [200, 201]:
            error(resp.reason)
        else:
            return json.loads(resp.read())

    def list_categories(self):
        data = self._get(f'{self.url}/category')['value']
        return data

    def create_category(self, category):
        pprint(f"Creating category {category}")
        create_spec = {"associable_types": ["VirtualMachine"], "cardinality": "MULTIPLE", "description": category,
                       "name": category}
        data = {'create_spec': create_spec}
        return self._post(f'{self.url}/category', data=data)['value']

    def delete_category(self, category):
        pprint(f"Deleting category {category}")
        category_id = self.get_category_id(category)
        self._delete(f'{self.url}/category/id:{category_id}')

    def get_category_id(self, category):
        if category.startswith('urn:vmomi'):
            return category
        for c in self.list_categories():
            resp = self._get(f'{self.url}/category/id:{c}')['value']
            if resp['name'] == category:
                return resp['id']

    def get_tag_id(self, tag):
        for t in self.list_tags():
            resp = self._get(f'{self.url}/tag/id:{t}')['value']
            if resp['name'] == tag:
                return resp['id']

    def get_tag_name(self, tag_id):
        for t in self.list_tags():
            resp = self._get(f'{self.url}/tag/id:{t}')['value']
            if resp['id'] == tag_id:
                return resp['name']

    def list_tags(self):
        data = self._get(f'{self.url}/tag')['value']
        return data

    def create_tag(self, category, tag):
        pprint(f"Creating tag {tag}")
        if category.startswith('urn:vmomi'):
            category_id = category
        else:
            category_id = self.get_category_id(category)
        create_spec = {"category_id": category_id, "description": tag, "name": tag, "tag_id": ""}
        data = {'create_spec': create_spec}
        return self._post(f'{self.url}/tag', data=data)['value']

    def delete_tag(self, tag):
        pprint(f"Deleting tag {tag}")
        tag_id = self.get_tag_id(tag)
        self._delete(f'{self.url}/tag/id:{tag_id}')

    def add_tag(self, vm_id, tag):
        pprint(f"Attaching tag {tag}")
        if tag.startswith('urn:vmomi'):
            tag_id = tag
        else:
            tag_id = self.get_tag_id(tag)
        object_id = {"id": vm_id, "type": "VirtualMachine"}
        data = {'object_id': object_id}
        return self._post(f'{self.url}/tag-association/id:{tag_id}?~action=attach', data=data)['value']

    def add_tags(self, vm_id, tag_ids):
        object_id = {"id": vm_id, "type": "VirtualMachine"}
        data = {'object_id': object_id, 'tag_ids': tag_ids}
        return self._post(f'{self.url}/tag-association?~action=attach-multiple-tags-to-object', data=data)

    def list_vm_tags(self, vm_id):
        tags = []
        object_id = {"id": vm_id, "type": "VirtualMachine"}
        data = {'object_id': object_id}
        for tag_id in self._post(f'{self.url}/tag-association?~action=list-attached-tags', data=data)['value']:
            tags.append(self.get_tag_name(tag_id))
        return tags



# coding=utf-8
from setuptools import setup, find_packages

import os
INSTALL = ['argcomplete', 'PyYAML', 'prettytable', 'jinja2', 'libvirt-python>=2.0.0']
AWS = ['boto3']
AZURE = ['azure-mgmt-compute', 'azure-mgmt-network', 'azure-mgmt-core', 'azure-identity', 'azure-mgmt-resource',
         'azure-mgmt-marketplaceordering', 'azure-storage-blob', 'azure-mgmt-dns', 'azure-mgmt-containerservice',
         'azure-mgmt-storage', 'azure-mgmt-msi', 'azure-mgmt-authorization']
HCLOUD = ['hcloud']
GCP = ['google-api-python-client', 'google-auth-httplib2', 'google-cloud-dns', 'google-cloud-storage',
       'google-cloud-container', 'google-cloud-compute']
OPENSTACK = ['python-cinderclient', 'python-neutronclient', 'python-glanceclient', 'python-keystoneclient',
             'python-novaclient', 'python-swiftclient']
OVIRT = ['ovirt-engine-sdk-python']
PROXMOX = ['proxmoxer']
VSPHERE = ['pyvmomi', 'cryptography']
IBMCLOUD = ['google-crc32c==1.1.2', 'ibm_vpc', 'ibm-cos-sdk', 'ibm-platform-services', 'ibm-cloud-networking-services']
EXTRAS = ['pyghmi', 'podman', 'websockify']
ALL = EXTRAS + AWS + GCP + OPENSTACK + OVIRT + VSPHERE + IBMCLOUD + AZURE + HCLOUD + PROXMOX

description = 'Provisioner/Manager for Libvirt/Vsphere/Aws/Gcp/Hcloud/Kubevirt/Ovirt/Openstack/IBM Cloud and containers'
long_description = description
if os.path.exists('README.rst'):
    long_description = open('README.rst').read()

setup(
    name='kcli',
    version='99.0.202506132125',
    include_package_data=True,
    packages=find_packages(),
    zip_safe=False,
    description=description,
    long_description=long_description,
    url='http://github.com/karmab/kcli',
    author='Karim Boumedhel',
    author_email='karimboumedhel@gmail.com',
    license='ASL',
    install_requires=INSTALL,
    extras_require={
        'all': ALL,
        'libvirt': [],
        'aws': AWS,
        'azure': AZURE,
        'gcp': GCP,
        'hcloud': HCLOUD,
        'ibm': IBMCLOUD,
        'openstack': OPENSTACK,
        'ovirt': OVIRT,
        'proxmox': PROXMOX,
        'vsphere': VSPHERE
    },
    entry_points='''
        [console_scripts]
        kcli=kvirt.cli:cli
        kweb=kvirt.web.main:run
        klist.py=kvirt.klist:main
        kmcpcore.py=kvirt.mcpcore:main
        kmcpcloud.py=kvirt.mcpcloud:main
        kmcpbm.py=kvirt.mcpbm:main
        ksushy=kvirt.ksushy.main:run
        ignitionmerger=kvirt.ignitionmerger:cli
        ekstoken=kvirt.ekstoken:cli
        gketoken=kvirt.gketoken:cli
    ''',
)



# -*- coding: utf-8 -*-

import os
from shutil import which


class Kcontainer():
    def __init__(self, host='127.0.0.1', user='root', port=22, engine='podman', debug=False, insecure=False):
        self.host = host
        self.user = user
        self.port = port
        self.debug = debug
        self.containermode = False
        self.engine = engine
        self.insecure = insecure
        if self.host == '127.0.0.1' and which(engine) is None:
            reason = f"executable {engine} not found in path"
            return {'result': 'failure', 'reason': reason}
        if self.host == '127.0.0.1' and os.path.exists("/i_am_a_container"):
            self.containermode = True
            if not os.path.exists("/run/podman/io.podman"):
                reason = "You need to add -v /run/podman/io.podman:/run/podman/io.podman to container alias"
                return {'result': 'failure', 'reason': reason}

    def create_container(self, name, image, nets=None, cmds=[], ports=[], volumes=[], environment=[], label=None,
                         overrides={}):
        engine = self.engine
        if self.containermode:
            finalvolumes = {}
            if volumes is not None:
                for i, volume in enumerate(volumes):
                    if isinstance(volume, str):
                        if len(volume.split(':')) == 2:
                            origin, destination = volume.split(':')
                            finalvolumes[origin] = {'bind': destination, 'mode': 'rw'}
                        else:
                            finalvolumes[volume] = {'bind': volume, 'mode': 'rw'}
                    elif isinstance(volume, dict):
                        path = volume.get('path')
                        origin = volume.get('origin')
                        destination = volume.get('destination')
                        mode = volume.get('mode', 'rw')
                        if origin is None or destination is None:
                            if path is None:
                                continue
                            finalvolumes[origin] = {'bind': path, 'mode': mode}
                        else:
                            finalvolumes[origin] = {'bind': destination, 'mode': mode}
            if ports is not None:
                ports = {f'{k}/tcp': k for k in ports}
            if label is not None and isinstance(label, str) and len(label.split('=')) == 2:
                key, value = label.split('=')
                labels = {key: value}
            else:
                labels = {'plan': 'kvirt'}
            finalenv = {}
            if environment is not None:
                for env in enumerate(environment):
                    if isinstance(env, str):
                        if len(env.split(':')) == 2:
                            key, value = env.split(':')
                            finalenv[key] = value
                        else:
                            continue
                    elif isinstance(env, dict):
                        if len(list(env)) == 1:
                            key = env.keys[0]
                            value = env[key]
                            finalenv[key] = value
                        else:
                            continue
            if ':' not in image:
                image = f'{image}:latest'
            cmd = ';'.join(cmds) if cmds else None
            self.d.containers.run(image, name=name, command=cmd, detach=True, ports=ports, volumes=finalvolumes,
                                  stdin_open=True, tty=True, labels=labels, environment=finalenv, stdout=True)
        else:
            portinfo = ''
            if ports is not None:
                for port in ports:
                    if isinstance(port, int):
                        oriport = port
                        destport = port
                    elif isinstance(port, str):
                        if len(port.split(':')) == 2:
                            oriport, destport = port.split(':')
                        else:
                            oriport = port
                            destport = port
                    elif isinstance(port, dict) and 'origin' in port and 'destination' in port:
                        oriport = port['origin']
                        destport = port['destination']
                    else:
                        continue
                    portinfo += f" -p {oriport}:{destport}"
            volumeinfo = ''
            if volumes is not None:
                for volume in volumes:
                    if isinstance(volume, str):
                        if len(volume.split(':')) == 2:
                            origin, destination = volume.split(':')
                        else:
                            origin = volume
                            destination = volume
                    elif isinstance(volume, dict):
                        path = volume.get('path')
                        origin = volume.get('origin')
                        destination = volume.get('destination')
                        if origin is None or destination is None:
                            if path is None:
                                continue
                            origin = path
                            destination = path
                    volumeinfo += f" -v {origin}:{destination}"
            envinfo = ''
            if environment is not None:
                for env in environment:
                    if isinstance(env, str):
                        if len(env.split(':')) == 2:
                            key, value = env.split(':')
                        else:
                            continue
                    elif isinstance(env, dict):
                        if len(list(env)) == 1:
                            key = list(env)[0]
                            value = env[key]
                        else:
                            continue
                    envinfo += f" -e {key}={value}"
            labelcmd = '-l plan=kvirt'
            if 'labels' in overrides:
                labels = overrides['labels']
                labelcmd = ' '.join(f"-l {label.split('=')[0]}={label.split('=')[1]}" for label in labels)
            selinux = "--security-opt label=disable" if self.insecure else ""
            runcommand = f"{engine} run {selinux} -it {volumeinfo} {envinfo} {portinfo} --name {name} {labelcmd}"
            runcommand += f" -d {image}"
            if cmds:
                runcommand += ' sh -c \\"%s\\"' % ';'.join(cmds)
            if self.host != '127.0.0.1':
                runcommand = f'ssh -p {self.port} {self.user}@{self.host} "{runcommand}"'
            if self.debug:
                print(runcommand)
            os.system(runcommand)
        return {'result': 'success'}

    def delete_container(self, name):
        engine = self.engine
        if self.containermode:
            containers = [container for container in self.d.containers.list() if container.name == name]
            if containers:
                for container in containers:
                    container.remove(force=True)
        else:
            rmcommand = f"{engine} rm -f {name}"
            if self.host != '127.0.0.1':
                rmcommand = f"ssh -p {self.port} {self.user}@{self.host} {rmcommand}"
            os.system(rmcommand)
        return {'result': 'success'}

    def start_container(self, name):
        engine = self.engine
        if self.containermode:
            containers = [container for container in self.d.containers.list(all=True) if container.name == name]
            if containers:
                for container in containers:
                    container.start()
        else:
            startcommand = f"{engine} start {name}"
            if self.host != '127.0.0.1':
                startcommand = f"ssh -p {self.port} {self.user}@{self.host} {startcommand}"
            os.system(startcommand)
        return {'result': 'success'}

    def stop_container(self, name):
        engine = self.engine
        if self.containermode:
            containers = [container for container in self.d.containers.list() if container.name == name]
            if containers:
                for container in containers:
                    container.stop()
        else:
            stopcommand = f"{engine} stop {name}"
            if self.host != '127.0.0.1':
                stopcommand = f"ssh -p {self.port} {self.user}@{self.host} {stopcommand}"
            os.system(stopcommand)
        return {'result': 'success'}

    def console_container(self, name):
        engine = self.engine
        if self.containermode:
            attachcommand = f"{engine} attach {name}"
            os.system(attachcommand)
        else:
            attachcommand = f"{engine} attach {name}"
            if self.host != '127.0.0.1':
                attachcommand = f"ssh -t -p {self.port} {self.user}@{self.host} {attachcommand}"
            os.system(attachcommand)
        return {'result': 'success'}

    def list_containers(self):
        engine = self.engine
        containers = []
        if self.containermode:
            for container in self.d.containers.list(all=True):
                name = container.name
                state = container.status
                state = state.split(' ')[0]
                if state.startswith('running'):
                    state = 'up'
                else:
                    state = 'down'
                source = container.attrs['Config']['Image']
                labels = container.attrs['Config']['Labels']
                if 'plan' in labels:
                    plan = labels['plan']
                else:
                    plan = 'kvirt'
                command = container.attrs['Config']['Cmd']
                if command is None:
                    command = ''
                else:
                    command = command[0]
                ports = container.attrs['NetworkSettings']['Ports']
                if ports:
                    portinfo = []
                    for port in ports:
                        if ports[port] is None:
                            newport = port
                        else:
                            hostport = ports[port][0]['HostPort']
                            hostip = ports[port][0]['HostIp']
                            newport = f"{hostip}:{hostport}->{port}"
                        portinfo.append(newport)
                    portinfo = ','.join(portinfo)
                else:
                    portinfo = ''
                containers.append([name, state, source, plan, command, portinfo, ''])
        else:
            containers = []
            lscommand = "%s ps -a --format \"{{.Names}}?{{.Status}}?{{.Image}}?{{.Command}}?{{.Ports}}?" % engine
            lscommand += "{{.Labels}}'\""
            if self.debug:
                print(lscommand)
            if self.host != '127.0.0.1':
                lscommand = f"ssh -p {self.port} {self.user}@{self.host} {lscommand}"
            results = os.popen(lscommand).readlines()
            for container in results:
                name, state, source, command, ports, labels = container.split('?')
                labels = labels.replace('map[', '').replace(']', '').replace('\n', '').split(' ')
                plan = 'kvirt'
                for label in labels:
                    if ':' not in label:
                        break
                    label_split = label.split(':')
                    key, value = label_split[0], ':'.join(label_split[1:])
                    if key == 'plan':
                        plan = value
                        break
                if state.startswith('Up'):
                    state = 'up'
                else:
                    state = 'down'
                command = command.strip().replace('"', '')
                containers.append([name, state, source, plan, command, ports, ''])
        return containers

    def exists_container(self, name):
        engine = self.engine
        if self.containermode:
            containers = [container.id for container in self.d.containers.list(all=True) if container.name == name]
            if containers:
                return True
        else:
            existcommand = "%s ps -a --format '{{.Names}}'" % engine
            if self.host != '127.0.0.1':
                existcommand = f"ssh -p {self.port} {self.user}@{self.host} {existcommand}"
            results = os.popen(existcommand).readlines()
            for container in results:
                containername = container.strip()
                if containername == name:
                    return True
        return False

    def list_images(self):
        engine = self.engine
        if self.containermode:
            images = []
            for i in self.d.images.list():
                for tag in i.tags:
                    images.append(tag)
        else:
            lsicommand = "%s images --format '{{.Repository}}:{{.Tag}}'" % engine
            if self.host != '127.0.0.1':
                lsicommand = f"ssh -p {self.port} {self.user}@{self.host} {lsicommand}"
            images = [image.strip() for image in os.popen(lsicommand).readlines()]
        return sorted(images)



# -*- coding: utf-8 -*-

from kvirt.common import error
from kvirt.kubecommon import _create_resource, _delete_resource, _get_resource, _get_all_resources
from kubernetes import client
import os
from shutil import which
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class Kubernetes():
    def __init__(self, kubeconfig_file, host='127.0.0.1', context=None,
                 namespace='default', readwritemany=False, debug=False, insecure=False):
        self.namespace = namespace or 'default'
        self.kubectl = which('kubectl') or which('oc')
        if self.kubectl is None:
            error("Kubectl is required. Use kcli download kubectl and put in your path")
            self.conn = None
            return
        elif context is not None:
            self.kubectl += f' --context {context}'
        os.environ['KUBECONFIG'] = os.path.expanduser(kubeconfig_file)
        self.host = host
        self.debug = debug
        self.insecure = insecure

    def create_container(self, name, image, nets=None, cmds=[], ports=[], volumes=[], environment=[], label=None,
                         overrides={}):
        """
        :param self:
        :param name:
        :param image:
        :param nets:
        :param cmds:
        :param ports:
        :param volumes:
        :param environment:
        :param label:
        :param overrides:
        :return:
        """
        kubectl = self.kubectl
        namespace = self.namespace
        if ':' not in image:
            image = f'{image}:latest'
        replicas = overrides.get('replicas', 1)
        if label is not None:
            if isinstance(label, str) and len(label.split('=')) == 2:
                key, value = label.split('=')
                labels = {key: value}
            elif isinstance(label, dict):
                labels = label
        else:
            labels = {}
        labels['kcli/plan'] = overrides.get('plan', 'kvirt')
        containers = [{'image': image, 'name': name, 'imagePullPolicy': 'IfNotPresent', 'ports': ports}]
        if cmds:
            containers[0]['command'] = cmds
        if volumes:
            vols = []
            containers[0]['volumeMounts'] = []
            for i, volume in enumerate(volumes):
                if isinstance(volume, str):
                    if len(volume.split(':')) == 2:
                        origin, destination = volume.split(':')
                        # mode = 'rw'
                    else:
                        origin, destination = volume, volume
                        # mode = 'rw'
                elif isinstance(volume, dict):
                    path = volume.get('path')
                    origin = volume.get('origin')
                    destination = volume.get('destination')
                    # mode = volume.get('mode', 'rw')
                    if (origin is None or destination is None) and path is None:
                        continue
                newvolmount = {'mountPath': destination, 'name': origin}
                containers[0]['volumeMounts'].append(newvolmount)
                newvol = {'name': origin, 'persistentVolumeClaim': {'claimName': origin}}
                vols.append(newvol)
        if ports:
            finalports = []
            for port in ports:
                finalports.append({'containerPort': port, 'name': f'port-{port}', 'protocol': 'TCP'})
            ports = finalports
        if environment:
            containers[0]['env'] = []
            for env in enumerate(environment):
                if isinstance(env, str):
                    if len(env.split(':')) == 2:
                        key, value = env.split(':')
                    else:
                        continue
                elif isinstance(env, dict):
                    if len(list(env)) == 1:
                        key = env.keys[0]
                        value = env[key]
                    else:
                        continue
                newenv = {'name': key, 'value': value}
                containers[0]['env'].append(newenv)
        deploy = {'apiVersion': 'extensions/v1beta1', 'kind': 'Deployment', 'metadata': {'labels': labels,
                                                                                         'name': name,
                                                                                         'namespace': self.namespace},
                  'spec': {'replicas': replicas, 'selector': {'matchLabels': labels},
                           'strategy': {'rollingUpdate': {'maxSurge': 1, 'maxUnavailable': 1}, 'type': 'RollingUpdate'},
                           'template': {'metadata': {'labels': labels}, 'spec': {'containers': containers,
                                                                                 'dnsPolicy': 'ClusterFirst',
                                                                                 'restartPolicy': 'Always'}}}}
        if volumes:
            deploy['spec']['template']['spec']['volumes'] = vols
        _create_resource(kubectl, deploy, namespace)

        return {'result': 'success'}

    def delete_container(self, name):
        kubectl = self.kubectl
        namespace = self.namespace
        try:
            pods = []
            rsname = None
            items = _get_all_resources(kubectl, 'rs', namespace)
            for rs in items:
                owner_references = rs['metadata']['ownerReferences']
                if owner_references is None:
                    continue
                ownerkind = owner_references[0]['kind']
                ownername = owner_references[0]['name']
                if ownerkind == 'Deployment' and ownername == name:
                    rsname = rs['metadata']['name']
                    items = _get_all_resources(kubectl, 'pod', namespace)
                    for pod in items:
                        owner_references = pod['metadata']['ownerReferences']
                        if owner_references is None:
                            continue
                        ownerkind = owner_references[0]['kind']
                        ownername = owner_references[0]['name']
                        if ownerkind == 'ReplicaSet' and ownername == rsname:
                            pods.append(pod['metadata']['name'])
            _delete_resource(kubectl, 'deploy', name, namespace)
            if rsname is not None:
                _delete_resource(kubectl, 'rs', name, rs['metadata']['name'])
            for pod in pods:
                _delete_resource(kubectl, 'pod', pod, namespace)
        except client.rest.ApiException:
            try:
                _delete_resource(kubectl, 'pod', pod, namespace)
            except client.rest.ApiException:
                error(f"Container {name} not found")
                return {'result': 'failure', 'reason': "Missing template"}
        return {'result': 'success'}

    def start_container(self, name):
        return {'result': 'success'}

    def stop_container(self, name):
        kubectl = self.kubectl
        namespace = self.namespace
        _delete_resource(kubectl, 'pod', name, namespace)
        return {'result': 'success'}

    def console_container(self, name):
        kubectl = self.kubectl
        os.system(f"{kubectl} exec -it {name} /bin/sh")
        return {'result': 'success'}

    def list_containers(self):
        kubectl = self.kubectl
        namespace = self.namespace
        containers = []
        items = _get_all_resources(kubectl, 'pod', namespace)
        for pod in items:
            name = pod['metadata']['name']
            state = pod['status']['phase']
            source = pod['spec']['containers'][0].image
            plan = ''
            labels = pod['metadata']['labels']
            if labels is not None and 'kcli/plan' in labels:
                plan = pod['metadata']['labels']['kcli/plan']
            command = pod['spec']['containers'][0]['command']
            if command is not None:
                command = ' '.join(command)
            portinfo = pod['spec']['nodeName']
            owner_references = pod['metadata']['ownerReferences']
            deploy = ''
            if owner_references is not None:
                ownerkind = owner_references[0]['kind']
                ownername = owner_references[0]['name']
                if ownerkind == 'ReplicaSet':
                    rs = _get_resource(kubectl, 'rs', ownername, namespace)
                    owner_references = rs['metadata']['ownerReferences']
                    if owner_references is not None:
                        ownerkind = owner_references[0]['kind']
                        ownername = owner_references[0]['name']
                        if ownerkind == 'Deployment':
                            deploy = ownername
            containers.append([name, state, source, plan, command, portinfo, deploy])
        return containers

    def exists_container(self, name):
        kubectl = self.kubectl
        namespace = self.namespace
        items = _get_all_resources(kubectl, 'pod', namespace)
        for pod in items:
            if pod['metadata']['name'] == name:
                return True
        return False

    def list_images(self):
        images = []
        return sorted(images)



# -*- coding: utf-8 -*-

from kvirt import common

# general notes
# most functions should either return
# return {'result': 'success'}
# or
# return {'result': 'failure', 'reason': reason}
# for instance
# return {'result': 'failure', 'reason': "VM %s not found" % name}


# your base class __init__ needs to define conn attribute and set it to None
# when backend cannot be reached
# it should also set debug from the debug variable passed in kcli client
class Kbase(object):
    def __init__(self, host='127.0.0.1', port=None, user='root', debug=False):
        self.conn = 'base'

# should cleanly close your connection, if needed
    def close(self):
        print("not implemented")

    def exists(self, name):
        print("not implemented")

    def net_exists(self, name):
        print("not implemented")

    def disk_exists(self, pool, name):
        print("not implemented")

    def create(self, name, virttype=None, profile='', flavor=None, plan='kvirt', cpumodel='host-model',
               cpuflags=[], cpupinning=[], numcpus=2, memory=512, guestid='guestrhel764', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True, diskinterface='virtio', nets=['default'], iso=None,
               vnc=True, cloudinit=True, reserveip=False, reservedns=False, reservehost=False, start=True, keys=[],
               cmds=[], ips=None, netmasks=None, gateway=None, nested=True, dns=None, domain=None, tunnel=False,
               files=[], enableroot=True, overrides={}, tags=[], storemetadata=False, sharedfolders=[],
               cmdline=None, placement=[], autostart=False, cpuhotplug=False, memoryhotplug=False,
               numamode=None, numa=[], pcidevices=[], tpm=False, rng=False, metadata={}, securitygroups=[],
               vmuser=None, guestagent=True):
        print("not implemented")
        return {'result': 'success'}

    def start(self, name):
        print("not implemented")
        return {'result': 'success'}

    def stop(self, name, soft=False):
        print("not implemented")
        return {'result': 'success'}

    def create_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def delete_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def list_snapshots(self, base):
        print("not implemented")
        return []

    def revert_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def restart(self, name):
        return self.start(name)

    def info_host(self):
        print("not implemented")
        return {}

    def status(self, name):
        print("not implemented")

# should return a sorted list of name, state, ip, source, plan, profile
    def list(self):
        print("not implemented")
        return []

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        print("not implemented")

    def serialconsole(self, name, web=False):
        print("not implemented")

# should generate info in a dict and then pass it to
# print_info(yamlinfo, output=output, fields=fields, values=values)
# from kvirt.common where:
# yamlinfo is the dict
# with the following keys (you can omit the ones you want)
# name
# autostart
# plan
# profile
# image
# ip
# memory
# numcpus
# creationdate
# nets list  of
# {'device': device, 'mac': mac, 'net': network, 'type': network_type}
# disks list of
# {'device': device, 'size': disksize, 'format': diskformat,
# 'type': drivertype, 'path': path}
# snapshots list of {'snapshot': snapshot, current: current}
# fields should be split with fields.split(',')
    def info(self, name, output='plain', fields=[], values=False, vm=None, debug=False):
        print("not implemented")
        return {'result': 'success'}

# should return ip string
    def ip(self, name):
        print("not implemented")

# should return a list of available images, or isos ( if iso is set to True
    def volumes(self, iso=False):
        print("not implemented")
        return []

    def delete(self, name, snapshots=False, keep_disks=False):
        print("not implemented")
        return {'result': 'success'}

# should return dnsclient, domain for the given vm
    def dnsinfo(self, name):
        return None, None

    def clone(self, old, new, full=False, start=False):
        print("not implemented")

    def update_metadata(self, name, metatype, metavalue, append=False):
        print("not implemented")

    def update_memory(self, name, memory):
        print("not implemented")

    def update_cpus(self, name, numcpus):
        print("not implemented")

    def update_start(self, name, start=True):
        print("not implemented")

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)

    def update_iso(self, name, iso):
        print("not implemented")

    def update_flavor(self, name, flavor):
        print("Not implemented")
        return {'result': 'success'}

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        print("not implemented")

    def add_disk(self, name, size=1, pool=None, thin=True, image=None, shareable=False, existing=None,
                 interface='virtio', novm=False, overrides={}):
        print("not implemented")

    def delete_disk(self, name, diskname, pool=None, novm=False):
        print("not implemented")

# should return a dict of {'diskname': {'pool': poolname, 'path': name}}
    def list_disks(self):
        print("not implemented")
        return []

    def add_nic(self, name, network, model='virtio'):
        print("not implemented")

    def delete_nic(self, name, interface):
        print("not implemented")

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")

    def delete_image(self, image, pool=None):
        print("not implemented")
        return {'result': 'success'}

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        print("not implemented")
        return {'result': 'success'}

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_network(self, name=None, cidr=None, force=False):
        print("not implemented")
        return {'result': 'success'}

# should return a list of pools
    def list_pools(self):
        print("not implemented")
        return []

    def list_networks(self):
        print("not implemented")
        return {}

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        return networkinfo

    def info_subnet(self, name):
        print("not implemented")
        return {}

    def list_subnets(self):
        print("not implemented")
        return {}

    def delete_pool(self, name, full=False):
        print("not implemented")

    def network_ports(self, name):
        print("not implemented")

    def vm_ports(self, name):
        print("not implemented")
        return ['default']

    def get_pool_path(self, pool):
        print("not implemented")

# return a list of [name, numcpus, memory] for each flavor, if the platform has this concept
    def list_flavors(self):
        return []

# export the primary disk of the corresponding instance so it's available as a image
    def export(self, name, image=None):
        print("not implemented")

    def create_bucket(self, bucket, public=False):
        print("not implemented")

    def delete_bucket(self, bucket):
        print("not implemented")

    def delete_from_bucket(self, bucket, path):
        print("not implemented")

    def download_from_bucket(self, bucket, path):
        print("not implemented")

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        print("not implemented")

    def list_buckets(self):
        print("not implemented")
        return []

    def list_bucketfiles(self, bucket):
        print("not implemented")
        return []

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False, instanceid=None):
        print("not implemented")

    def update_nic(self, name, index, network):
        print("not implemented")

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_security_groups(self, network=None):
        print("not implemented")
        return []

    def create_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_security_group(self, name):
        print("not implemented")
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        print("not implemented")
        return []

    def detach_disks(self, name):
        print("not implemented")
        return {'result': 'success'}



import functools
from kvirt.bottle import Bottle, request, static_file, jinja2_view, response
from glob import glob
from kvirt.common import get_overrides, pprint, get_parameters
import os
import re
import yaml


def convert_value(value):
    if value.isdigit():
        value = int(value)
    elif value.lower() in ['true', 'false']:
        value = value.lower() == "true"
    return value


class Kexposer():

    def refresh_plans(self, verbose=False):
        plans = []
        owners = {}
        for paramfile in glob(f"{self.basedir}/**/parameters_*.y*ml", recursive=True):
            search = re.match('.*parameters_(.*)\\.ya?ml', paramfile)
            plan = search.group(1)
            try:
                fileoverrides = get_overrides(paramfile=paramfile)
            except:
                continue
            if 'owner' in fileoverrides:
                owners[plan] = fileoverrides['owner']
            plans.append(plan)
            if verbose:
                pprint(f"Adding parameter file {paramfile}")
        self.plans = sorted(plans) if plans else [self.plan]
        self.owners = owners

    def get_client(self, plan, currentconfig, overrides={}):
        matching = glob(f"{self.basedir}/**/parameters_{plan}.y*ml", recursive=True)
        if matching:
            paramfile = matching[0]
            fileoverrides = get_overrides(paramfile=paramfile)
            if 'client' in fileoverrides:
                client = fileoverrides['client']
                currentconfig.__init__(client=client)
                if overrides:
                    fileoverrides.update(overrides)
                    overrides = fileoverrides
        return overrides

    def __init__(self, config, plan, inputfile, overrides={}, port=9000, pfmode=False, cluster=False, extras=False):
        app = Bottle()
        basedir = f"{os.path.dirname(Bottle.run.__code__.co_filename)}/expose"
        view = functools.partial(jinja2_view, template_lookup=[f"{basedir}/templates"])
        self.basedir = os.path.dirname(inputfile) if '/' in inputfile else '.'
        self.plan = plan
        self.overrides = overrides
        self.pfmode = pfmode
        self.extras = extras
        self.cluster = cluster
        self.refresh_plans(verbose=True)
        self.inputfile = inputfile

        @app.route('/static/<filename:path>')
        def server_static(filename):
            return static_file(filename, root=f'{basedir}/static')

        @app.route('/')
        @view('index.html')
        def index():
            self.refresh_plans()
            return {'plans': self.plans, 'owners': self.owners}

        @app.route('/planstable')
        @view('planstable.html')
        def plansstable():
            self.refresh_plans()
            return {'plans': self.plans, 'owners': self.owners}

        @app.route("/exposecreate", method='POST')
        @view('result.html')
        def exposecreateweb():
            update = False
            currentconfig = self.config
            if 'plan' in request.forms:
                plan = request.forms['plan']
                if plan not in self.plans:
                    return f'Invalid plan name {plan}'
                pfdata = None
                parameters = {}
                for p in request.forms:
                    if p.startswith('parameter'):
                        value = request.forms[p]
                        if value.isdigit():
                            value = int(value)
                        elif value.lower() in ['true', 'false']:
                            value = value.lower() == "true"
                        key = p.replace('parameter_', '')
                        parameters[key] = value
                    elif p == 'update':
                        update = True
                    elif p == 'pf':
                        pfdata = request.forms[p]
                if pfdata is not None:
                    try:
                        new_parameters = yaml.safe_load(pfdata)
                    except:
                        new_parameters = {}
                    new_parameters.update(parameters)
                    parameters = new_parameters
                try:
                    overrides = self.get_client(plan, currentconfig, overrides=parameters)
                    if 'mail' in currentconfig.notifymethods and 'mail_to' in overrides and overrides['mail_to'] != "":
                        newmails = overrides['mail_to'].split(',')
                        if currentconfig.mailto:
                            currentconfig.mailto.extend(newmails)
                        else:
                            currentconfig.mailto = newmails
                    if 'owner' in overrides and overrides['owner'] == '':
                        del overrides['owner']
                    inputfile = overrides.get('inputfile', self.inputfile)
                    if update:
                        result = currentconfig.plan(plan, inputfile=inputfile, overrides=overrides, update=True)
                    else:
                        if self.cluster:
                            currentconfig.delete_kube(plan)
                        else:
                            currentconfig.delete_plan(plan)
                        result = currentconfig.plan(plan, inputfile=inputfile, overrides=overrides)
                except Exception as e:
                    error = f'Hit issue when running plan: {str(e)}'
                    response.status = 400
                    result = {'result': 'failure', 'error': error}
                result.update({'plan': plan})
                return result
            else:
                return 'Missing plan in data'

        @app.route("/exposeform/<plan>")
        @view('form.html')
        def exposeform(plan):
            parameters = self.overrides
            if plan not in self.plans:
                return f'Invalid plan name {plan}'
            inputfile = self.overrides.get('inputfile', self.inputfile)
            info = get_parameters(inputfile, planfile=True)
            info = info.get('info', 'N/A')
            return {'parameters': parameters, 'plan': plan, 'pfmode': self.pfmode, 'info': info, 'extras': self.extras}

        def _infoplan(plan):
            currentconfig = self.config
            if plan not in self.plans:
                return f'Invalid plan name {plan}'
            self.get_client(plan, currentconfig)
            vms = currentconfig.info_specific_plan(plan)
            creationdate = ''
            owner = self.owners[plan] if plan in self.owners else ''
            if vms:
                creationdate = vms[0].get('creationdate', creationdate)
                if 'owner' in vms[0]:
                    owner = vms[0]['owner']
                return {'vms': vms, 'plan': plan, 'client': currentconfig.client, 'creationdate': creationdate,
                        'owner': owner}

        def _infocluster(plan):
            currentconfig = self.config
            data = currentconfig.info_specific_kube(plan)
            data['kube'] = plan
            return data

        @app.route("/infoplan/<plan>")
        @view('infoplan.html')
        def infoplan(plan):
            if self.cluster:
                return _infocluster(plan)
            else:
                return _infoplan(plan)

        # API

        @app.route('/expose')
        def exposelist():
            return {'plans': self.plans, 'owners': self.owners, 'overrides': self.overrides}

        @app.route("/expose", method='POST')
        def exposeplancreate():
            update = False
            currentconfig = self.config
            json = request.json
            if json is None:
                response.status = 400
                return 'Invalid json'
            if 'plan' in json:
                plan = json['plan']
                if plan not in self.plans:
                    response.status = 400
                    return f'Invalid plan name {plan}'
                pfdata = None
                parameters = {}
                for p in json:
                    value = request.json[p]
                    if p == 'extras':
                        extra_parameters = {}
                        for entry in value.strip().split('-P '):
                            if entry != '':
                                new_param, new_value = entry.split('=')
                                extra_parameters[new_param] = convert_value(new_value)
                        parameters.update(extra_parameters)
                    elif p not in self.overrides:
                        print(f"Skipping parameter {p}")
                        continue
                    parameters[p] = convert_value(new_value)
                if pfdata is not None:
                    try:
                        new_parameters = yaml.safe_load(pfdata)
                    except:
                        new_parameters = {}
                    new_parameters.update(parameters)
                    parameters = new_parameters
                print(parameters)
                try:
                    overrides = self.get_client(plan, currentconfig, overrides=parameters)
                    if 'mail' in currentconfig.notifymethods and 'mail_to' in overrides and overrides['mail_to'] != "":
                        newmails = overrides['mail_to'].split(',')
                        if currentconfig.mailto:
                            currentconfig.mailto.extend(newmails)
                        else:
                            currentconfig.mailto = newmails
                    if 'owner' in overrides and overrides['owner'] == '':
                        del overrides['owner']
                    inputfile = self.overrides.get('inputfile', self.inputfile)
                    if update:
                        result = currentconfig.plan(plan, inputfile=inputfile, overrides=overrides, update=True)
                    else:
                        if self.cluster:
                            currentconfig.delete_kube(plan)
                        else:
                            currentconfig.delete_plan(plan)
                        result = currentconfig.plan(plan, inputfile=inputfile, overrides=overrides)
                except Exception as e:
                    error = f'Hit issue when running plan: {str(e)}'
                    response.status = 400
                    result = {'result': 'failure', 'error': error}
                return {'plan': plan, 'result': result}
            else:
                return 'Missing plan in data'

        @app.route("/expose/<plan>")
        def exposeplaninfo(plan):
            if self.cluster:
                return _infocluster(plan)
            else:
                return _infoplan(plan)

        @app.route('/expose/kubeconfig/kubeconfig.<kube>')
        def kubeconfigfile(kube):
            kubeconfig = os.path.expanduser(f'~/.kcli/clusters/{kube}/auth/kubeconfig')
            if self.cluster and os.path.exists(kubeconfig):
                return static_file('kubeconfig', root=os.path.dirname(kubeconfig), mimetype='application/octet-stream')
            else:
                response.status = 404
                return {}

        @app.route('/expose/<kube>/kubeconfig')
        def kubeconfig(kube):
            kubeconfig = os.path.expanduser(f'~/.kcli/clusters/{kube}/auth/kubeconfig')
            if self.cluster and os.path.exists(kubeconfig):
                return open(kubeconfig).read()
            else:
                response.status = 404
                return {}

        @app.route("/expose/<plan>", method=['DELETE'])
        def exposedelete(plan):
            currentconfig = self.config
            if plan not in self.plans:
                response.status = 400
                return f'Invalid plan name {plan}'
            self.get_client(plan, currentconfig)
            if self.cluster:
                result = currentconfig.delete_kube(plan)
            else:
                result = currentconfig.delete_plan(plan)
            response.status = 200
            return result

        self.app = app
        self.config = config
        self.overrides = overrides
        self.port = port

    def run(self):
        self.app.run(host='0.0.0.0', port=self.port)



from base64 import b64encode
from glob import glob
from ipaddress import ip_address, ip_network
import json
import os
import re
import sys
from urllib.request import urlopen
from time import sleep
import yaml


def basename(path):
    return os.path.basename(path)


def diskpath(path):
    return path if path.startswith('/dev') else f"/dev/{path}"


def dirname(path):
    return os.path.dirname(path)


def none(value):
    return value if value is not None else ''


def base64(value):
    if value is None:
        return None
    return str(b64encode(str(value).encode('utf-8')), 'utf-8')


def _type(value):
    if value is None:
        return None
    elif isinstance(value, str):
        return 'string'
    elif isinstance(value, int):
        return 'int'
    elif isinstance(value, dict):
        return 'dict'
    elif isinstance(value, list):
        return 'list'


def kubenodes(cluster, ctlplanes=1, workers=1):
    ctlplanes = [f'{cluster}-ctlplane-{num}' for num in range(ctlplanes)]
    workers = [f'{cluster}-worker-{num}' for num in range(workers)]
    return ctlplanes + workers


def certificate(value):
    if 'BEGIN CERTIFICATE' in value:
        return value
    else:
        return "-----BEGIN CERTIFICATE-----\n%s\n-----END CERTIFICATE-----" % value


def stable_release(release, tag_mode=False):
    name = 'name' if tag_mode else 'tag_name'
    tag = release[name]
    if 'rc' in tag or 'alpha' in tag or 'beta' in tag:
        return False
    if 'prerelease' in release and release['prerelease']:
        return False
    return True


def github_version(repo, version=None, tag_mode=False):
    if version is None or version == 'latest':
        obj = 'tags' if tag_mode else 'releases'
        tag_name = 'name' if tag_mode else 'tag_name'
        data = json.loads(urlopen(f"https://api.github.com/repos/{repo}/{obj}", timeout=5).read())
        if 'message' in data and data['message'] == 'Not Found':
            return ''
        tags = sorted([x[tag_name] for x in data if stable_release(x, tag_mode)],
                      key=lambda string: list(map(int, re.findall(r'\d+', string)))[0], reverse=True)
        if tags:
            tag = tags[0]
        else:
            tag = data[0][tag_name]
        print('\033[0;36mUsing version %s %s\033[0;0m' % (os.path.basename(repo), tag))
        return tag


def defaultnodes(replicas, cluster, domain, ctlplanes, workers):
    nodes = []
    for num in range(workers):
        if len(nodes) < replicas:
            nodes.append(f'{cluster}-worker-{num}.{domain}')
    for num in range(ctlplanes):
        if len(nodes) < replicas:
            nodes.append(f'{cluster}-ctlplane-{num}.{domain}')
    return nodes


def wait_crd(crd, timeout=120):
    result = """timeout=0
ready=false
while [ "$timeout" -lt "%s" ] ; do
  oc get crd | grep -q %s && ready=true && break;
  echo "Waiting for CRD %s to be created"
  sleep 5
  timeout=$(($timeout + 5))
done
if [ "$ready" == "false" ] ; then
 echo Timeout waiting for CRD %s
 exit 1
fi """ % (timeout, crd.lower(), crd, crd)
    return result


def wait_csv(csv, namespace, timeout=360):
    result = """timeout=0
ready=false
while [ "$timeout" -lt "%s" ] ; do
  [ "$(oc get csv -n %s %s -o jsonpath='{.status.phase}')" == "Succeeded" ] && ready=true && break;
  echo "Waiting for CSV %s to be created"
  sleep 5
  timeout=$(($timeout + 5))
done
if [ "$ready" == "false" ] ; then
 echo Timeout waiting for CSV %s
 exit 1
fi """ % (timeout, namespace, csv, csv, csv)
    return result


def local_ip(net, wrap=False):
    c = "ip a s %s 2>/dev/null | grep -E 'inet6?[[:space:]][^fe]' |head -1| awk '{print $2}' | cut -d '/' -f 1" % net
    result = os.popen(c).read().strip()
    if result == '' and net == 'default':
        c = "ip a s virbr0 2>/dev/null | grep -E 'inet6?[[:space:]][^fe]' |head -1| awk '{print $2}' | cut -d '/' -f 1"
        result = os.popen(c).read().strip()
    if wrap and ':' in result:
        result = '[%s]' % result
    return result


def network_ip(network, num=0, version=False):
    try:
        ip = str(ip_network(network)[num])
        if version and ':' in network:
            return f"[{ip}]"
        else:
            return ip
    except Exception as e:
        print(f"Error processing filter network_ip with {network} and {num}. Got {e}")
        sys.exit(1)


def kcli_info(name, key=None, client=None, wait=False):
    client_header = f'-C {client}' if client is not None else ''
    if key is not None:
        c = f"kcli {client_header} info vm -vf {key} {name}"
        result = os.popen(c).read().strip()
    else:
        c = f"kcli {client_header} info vm -o yaml {name}"
        result = yaml.load(os.popen(c).read())
    if result == '' and wait:
        sleep(10)
        print(f"Waiting 10s for info to be available on {name}")
        return kcli_info(name, key=key, client=client, wait=wait)
    return result


def find_manifests(directory, suffix='yaml'):
    results = []
    for f in glob("%s/*.y*ml" % directory):
        results.append(os.path.basename(f))
    return results


def exists(name):
    if name is None:
        return False
    return True if os.path.exists(name) else False


def ipv6_wrap(name):
    try:
        if ip_address(name).version == 6:
            return f'[{name}]'
        else:
            return name
    except:
        return name


def has_ctlplane(_list):
    for entry in _list:
        if 'ctlplane' in entry or 'master' in entry:
            return True
    return False


def count(var, char):
    return str(var).count(char)


def pwd_path(path):
    if path is not None and not os.path.isabs(path) and os.path.exists("/i_am_a_container")\
       and os.path.exists('/workdir'):
        return f'/workdir/{path}'
    else:
        return path


def real_path(path):
    return path and os.path.realpath(path)


def read_file(path):
    with open(path, 'r') as f:
        return f.read()


def min_ocp_version(version1, version2):
    return version1 if version1 < version2 else version2


def max_ocp_version(version1, version2):
    return version1 if version1 > version2 else version2


jinjafilters = {'basename': basename, 'dirname': dirname, 'diskpath': diskpath, 'kubenodes': kubenodes, 'none': none,
                'type': _type, 'certificate': certificate, 'base64': base64, 'github_version': github_version,
                'defaultnodes': defaultnodes, 'wait_crd': wait_crd, 'local_ip': local_ip, 'network_ip': network_ip,
                'kcli_info': kcli_info, 'find_manifests': find_manifests, 'exists': exists, 'ipv6_wrap': ipv6_wrap,
                'has_ctlplane': has_ctlplane, 'wait_csv': wait_csv, 'count': count, 'pwd_path': pwd_path,
                'min_ocp_version': min_ocp_version, 'max_ocp_version': max_ocp_version, 'real_path': real_path,
                'read_file': read_file}


class FilterModule(object):
    def filters(self):
        return jinjafilters



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random

left = ["astonishing", "happy", "jolly", "dreamy", "sad", "angry", "pensive", "focused", "sleepy", "grave",
        "distracted", "determined", "stoic", "stupefied", "sharp", "agitated", "cocky", "tender", "goofy", "furious",
        "desperate", "hopeful", "compassionate", "silly", "lonely", "condescending", "naughty", "kickass", "drunk",
        "boring", "nostalgic", "ecstatic", "insane", "cranky", "mad", "jovial", "sick", "hungry", "thirsty",
        "elegant", "backstabbing", "clever", "trusting", "loving", "suspicious", "berserk", "high", "romantic",
        "prickly", "evil", "magic", "vaccinated", "blader", "starving", "dope", "dingo", "summer", "tolai", "sweaty",
        "fired"]

right = ["albattani", "ablonde", "acid", "adark", "adebarre", "almeida", "alcaraz", "acosta", "alknopfler", "amazigh",
         "amouranth", "afawx", "archimedes", "ardinghelli", "babbage", "bagnaia", "bardeen", "bartik", "bcoco",
         "benzema", "bingchong", "bbanks", "bell", "beyonder", "blackwell", "bohr", "brattain", "briana", "brown",
         "bumblefoot", "cantalpino", "carson", "cell", "chechik", "clapton", "cobra", "colden", "connors", "conti",
         "cr7", "curie", "darwin", "davinci", "django", "dolores", "einstein", "elion", "engelbart", "euclid", "fermat",
         "fermi", "feynman", "franklin", "freezer", "galactus", "galois", "galileo", "ghowe", "giselle", "gogetta",
         "gohan", "goku", "goldstine", "goodall", "govan", "griezmann", "hawking", "heisenberg",
         "hoover", "hopper", "hypatia", "idir", "jhendrix", "jmayer", "jmiller", "jonas", "jones", "kirch", "kendra",
         "kheiron", "kowalevski", "krilin", "kylee", "lalande", "leakey", "lendl", "lilou", "lomax", "lovelace",
         "lumiere", "lomax", "macenroe", "mwilde", "marquez", "martin",
         "chavero", "matoub", "jmayer", "mbappe", "mccarthy", "mcclintock", "mclean", "meitner", "messi", "mestorf",
         "mikel", "vaiana", "stevemorse", "nilibrosh", "newton", "nobel", "nuno", "norinradd", "obama", "pare",
         "pasteur", "perlman", "picolo", "pike", "pinkman", "poincare", "pparker", "productionready", "ptolemy",
         "quartararo", "rosalind", "rgemma", "reinhardt", "rossi", "sammet", "sandler", "satriani", "selen",
         "shockley", "shyla", "sinoussi", "silversurfer", "snowwhites", "socarrat", "spidey", "spitzer", "stallman",
         "stochelo", "syren", "tesla", "thompson", "torvalds", "turing", "wilson", "wozniak", "wright",
         "yonath", "tudela", "tylerdurdeen", "korsani", "karmab", "gotrunks", "xhamster", "djokovic", "federer",
         "nadal", "sampras", "valadas", "vai", "vegeta", "ivanisevic", "muster", "agassi", "targaryen", "lannister",
         "safin", "stark", "jsnow", "sinner", "superduper", "tommyemmanuel", "thor", "mario", "walterwhite", "yuval",
         "5g", "5guys", "200k", "merletravis", "chetatkins", "fusilijerri", "joerobinson", "mog", "unity", "savannah",
         "soul", "mm93", "vcossart", "gruson", "patarin", "hichem", "regis", "elisa", "brais", "duncanmcleod", "mizu",
         "makio", "lois", "fender", "marshall", "stratocaster", "telecaster", "tenshinhan", "jmartin", "dalridge",
         "paul", "john", "mellen", "johnwick", "stef", "sperez", "sweetstance", "kindgrind", "fishbrain", "topsoul",
         "alleyoop", "mancuso", "asato", "atimmons", "interstellar", "shawshank", "samourai", "neo", "morfeus"]


def get_random_name(sep='-'):
    r = random.SystemRandom()
    while 1:
        name = '%s%s%s' % (r.choice(left), sep, r.choice(right))
        return name


def random_ip():
    ip = ".".join(map(str, (random.randint(0, 255) for _ in range(4))))
    return ip


if __name__ == '__main__':
    print(get_random_name())
